# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib
import http.client
import json
from unittest import mock
import urllib
import uuid

from keystoneclient.contrib.ec2 import utils as ec2_utils
from oslo_db import exception as oslo_db_exception
from testtools import matchers

from keystone.api import ec2tokens
from keystone.common import provider_api
from keystone.common import utils
from keystone.credential.providers import fernet as credential_fernet
from keystone import exception
from keystone import oauth1
from keystone.tests import unit
from keystone.tests.unit import ksfixtures
from keystone.tests.unit import test_v3

PROVIDERS = provider_api.ProviderAPIs
CRED_TYPE_EC2 = ec2tokens.CRED_TYPE_EC2


class CredentialBaseTestCase(test_v3.RestfulTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'credential',
                credential_fernet.MAX_ACTIVE_KEYS,
            )
        )

    def _create_dict_blob_credential(self):
        blob, credential = unit.new_ec2_credential(
            user_id=self.user['id'], project_id=self.project_id
        )

        # Store the blob as a dict *not* JSON ref bug #1259584
        # This means we can test the dict->json workaround, added
        # as part of the bugfix for backwards compatibility works.
        credential['blob'] = blob
        credential_id = credential['id']

        # Create direct via the DB API to avoid validation failure
        PROVIDERS.credential_api.create_credential(credential_id, credential)

        return json.dumps(blob), credential_id

    def _test_get_token(self, access, secret):
        """Test signature validation with the access/secret provided."""
        signer = ec2_utils.Ec2Signer(secret)
        params = {
            'SignatureMethod': 'HmacSHA256',
            'SignatureVersion': '2',
            'AWSAccessKeyId': access,
        }
        request = {
            'host': 'foo',
            'verb': 'GET',
            'path': '/bar',
            'params': params,
        }
        signature = signer.generate(request)

        # Now make a request to validate the signed dummy request via the
        # ec2tokens API.  This proves the v3 ec2 credentials actually work.
        sig_ref = {
            'access': access,
            'signature': signature,
            'host': 'foo',
            'verb': 'GET',
            'path': '/bar',
            'params': params,
        }
        r = self.post(
            '/ec2tokens',
            body={'ec2Credentials': sig_ref},
            expected_status=http.client.OK,
        )
        self.assertValidTokenResponse(r)
        return r.result['token']


class CredentialTestCase(CredentialBaseTestCase):
    """Test credential CRUD."""

    def setUp(self):
        super().setUp()

        self.credential = unit.new_credential_ref(
            user_id=self.user['id'], project_id=self.project_id
        )

        PROVIDERS.credential_api.create_credential(
            self.credential['id'], self.credential
        )

    def test_credential_api_delete_credentials_for_project(self):
        PROVIDERS.credential_api.delete_credentials_for_project(
            self.project_id
        )
        # Test that the credential that we created in .setUp no longer exists
        # once we delete all credentials for self.project_id
        self.assertRaises(
            exception.CredentialNotFound,
            PROVIDERS.credential_api.get_credential,
            credential_id=self.credential['id'],
        )

    def test_credential_api_delete_credentials_for_user(self):
        PROVIDERS.credential_api.delete_credentials_for_user(self.user_id)
        # Test that the credential that we created in .setUp no longer exists
        # once we delete all credentials for self.user_id
        self.assertRaises(
            exception.CredentialNotFound,
            PROVIDERS.credential_api.get_credential,
            credential_id=self.credential['id'],
        )

    def test_list_credentials(self):
        """Call ``GET /credentials``."""
        r = self.get('/credentials')
        self.assertValidCredentialListResponse(r, ref=self.credential)

    def test_list_credentials_filtered_by_user_id(self):
        """Call ``GET  /credentials?user_id={user_id}``."""
        credential = unit.new_credential_ref(user_id=uuid.uuid4().hex)
        PROVIDERS.credential_api.create_credential(
            credential['id'], credential
        )

        r = self.get('/credentials?user_id={}'.format(self.user['id']))
        self.assertValidCredentialListResponse(r, ref=self.credential)
        for cred in r.result['credentials']:
            self.assertEqual(self.user['id'], cred['user_id'])

    def test_list_credentials_filtered_by_type(self):
        """Call ``GET  /credentials?type={type}``."""
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.role_id
        )
        token = self.get_system_scoped_token()

        # The type ec2 was chosen, instead of a random string,
        # because the type must be in the list of supported types
        ec2_credential = unit.new_credential_ref(
            user_id=uuid.uuid4().hex,
            project_id=self.project_id,
            type=CRED_TYPE_EC2,
        )

        ec2_resp = PROVIDERS.credential_api.create_credential(
            ec2_credential['id'], ec2_credential
        )

        # The type cert was chosen for the same reason as ec2
        r = self.get('/credentials?type=cert', token=token)

        # Testing the filter for two different types
        self.assertValidCredentialListResponse(r, ref=self.credential)
        for cred in r.result['credentials']:
            self.assertEqual('cert', cred['type'])

        r_ec2 = self.get('/credentials?type=ec2', token=token)
        self.assertThat(r_ec2.result['credentials'], matchers.HasLength(1))
        cred_ec2 = r_ec2.result['credentials'][0]

        self.assertValidCredentialListResponse(r_ec2, ref=ec2_resp)
        self.assertEqual(CRED_TYPE_EC2, cred_ec2['type'])
        self.assertEqual(ec2_credential['id'], cred_ec2['id'])

    def test_list_credentials_filtered_by_type_and_user_id(self):
        """Call ``GET  /credentials?user_id={user_id}&type={type}``."""
        user1_id = uuid.uuid4().hex
        user2_id = uuid.uuid4().hex

        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.role_id
        )
        token = self.get_system_scoped_token()

        # Creating credentials for two different users
        credential_user1_ec2 = unit.new_credential_ref(
            user_id=user1_id, type=CRED_TYPE_EC2
        )
        credential_user1_cert = unit.new_credential_ref(user_id=user1_id)
        credential_user2_cert = unit.new_credential_ref(user_id=user2_id)

        PROVIDERS.credential_api.create_credential(
            credential_user1_ec2['id'], credential_user1_ec2
        )
        PROVIDERS.credential_api.create_credential(
            credential_user1_cert['id'], credential_user1_cert
        )
        PROVIDERS.credential_api.create_credential(
            credential_user2_cert['id'], credential_user2_cert
        )

        r = self.get(f'/credentials?user_id={user1_id}&type=ec2', token=token)
        self.assertValidCredentialListResponse(r, ref=credential_user1_ec2)
        self.assertThat(r.result['credentials'], matchers.HasLength(1))
        cred = r.result['credentials'][0]
        self.assertEqual(CRED_TYPE_EC2, cred['type'])
        self.assertEqual(user1_id, cred['user_id'])

    def test_create_credential(self):
        """Call ``POST /credentials``."""
        ref = unit.new_credential_ref(user_id=self.user['id'])
        r = self.post('/credentials', body={'credential': ref})
        self.assertValidCredentialResponse(r, ref)

    def test_get_credential(self):
        """Call ``GET /credentials/{credential_id}``."""
        r = self.get(
            '/credentials/{credential_id}'.format(
                credential_id=self.credential['id']
            )
        )
        self.assertValidCredentialResponse(r, self.credential)

    def test_update_credential(self):
        """Call ``PATCH /credentials/{credential_id}``."""
        ref = unit.new_credential_ref(
            user_id=self.user['id'], project_id=self.project_id
        )
        del ref['id']
        r = self.patch(
            '/credentials/{credential_id}'.format(
                credential_id=self.credential['id']
            ),
            body={'credential': ref},
        )
        self.assertValidCredentialResponse(r, ref)

    def test_update_credential_to_ec2_type(self):
        """Call ``PATCH /credentials/{credential_id}``."""
        # Create a credential without providing a project_id
        ref = unit.new_credential_ref(user_id=self.user['id'])
        r = self.post('/credentials', body={'credential': ref})
        self.assertValidCredentialResponse(r, ref)
        credential_id = r.result.get('credential')['id']

        # Updating the credential to ec2 requires a project_id
        update_ref = {'type': 'ec2', 'project_id': self.project_id}
        self.patch(
            f'/credentials/{credential_id}', body={'credential': update_ref}
        )

    def test_update_credential_to_ec2_missing_project_id(self):
        """Call ``PATCH /credentials/{credential_id}``."""
        # Create a credential without providing a project_id
        ref = unit.new_credential_ref(user_id=self.user['id'])
        r = self.post('/credentials', body={'credential': ref})
        self.assertValidCredentialResponse(r, ref)
        credential_id = r.result.get('credential')['id']

        # Updating such credential to ec2 type without providing a project_id
        # will fail
        update_ref = {'type': 'ec2'}
        self.patch(
            f'/credentials/{credential_id}',
            body={'credential': update_ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_update_credential_to_ec2_with_previously_set_project_id(self):
        """Call ``PATCH /credentials/{credential_id}``."""
        # Create a credential providing a project_id
        ref = unit.new_credential_ref(
            user_id=self.user['id'], project_id=self.project_id
        )
        r = self.post('/credentials', body={'credential': ref})
        self.assertValidCredentialResponse(r, ref)
        credential_id = r.result.get('credential')['id']

        # Since the created credential above already has a project_id, the
        # update request will not fail
        update_ref = {'type': 'ec2'}
        self.patch(
            f'/credentials/{credential_id}', body={'credential': update_ref}
        )

    def test_update_credential_non_owner(self):
        """Call ``PATCH /credentials/{credential_id}``."""
        alt_user = unit.create_user(
            PROVIDERS.identity_api, domain_id=self.domain_id
        )
        alt_user_id = alt_user['id']
        alt_project = unit.new_project_ref(domain_id=self.domain_id)
        alt_project_id = alt_project['id']
        PROVIDERS.resource_api.create_project(alt_project['id'], alt_project)
        alt_role = unit.new_role_ref(name='reader')
        alt_role_id = alt_role['id']
        PROVIDERS.role_api.create_role(alt_role_id, alt_role)
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            alt_user_id, alt_project_id, alt_role_id
        )
        auth = self.build_authentication_request(
            user_id=alt_user_id,
            password=alt_user['password'],
            project_id=alt_project_id,
        )
        ref = unit.new_credential_ref(
            user_id=alt_user_id, project_id=alt_project_id
        )
        r = self.post('/credentials', auth=auth, body={'credential': ref})
        self.assertValidCredentialResponse(r, ref)
        credential_id = r.result.get('credential')['id']

        # Cannot change the credential to be owned by another user
        update_ref = {'user_id': self.user_id, 'project_id': self.project_id}
        self.patch(
            f'/credentials/{credential_id}',
            expected_status=403,
            auth=auth,
            body={'credential': update_ref},
        )

    def test_update_ec2_credential_change_trust_id(self):
        """Call ``PATCH /credentials/{credential_id}``."""
        blob, ref = unit.new_ec2_credential(
            user_id=self.user['id'], project_id=self.project_id
        )
        blob['trust_id'] = uuid.uuid4().hex
        ref['blob'] = json.dumps(blob)
        r = self.post('/credentials', body={'credential': ref})
        self.assertValidCredentialResponse(r, ref)
        credential_id = r.result.get('credential')['id']
        # Try changing to a different trust
        blob['trust_id'] = uuid.uuid4().hex
        update_ref = {'blob': json.dumps(blob)}
        self.patch(
            f'/credentials/{credential_id}',
            body={'credential': update_ref},
            expected_status=http.client.BAD_REQUEST,
        )
        # Try removing the trust
        del blob['trust_id']
        update_ref = {'blob': json.dumps(blob)}
        self.patch(
            f'/credentials/{credential_id}',
            body={'credential': update_ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_update_ec2_credential_change_app_cred_id(self):
        """Call ``PATCH /credentials/{credential_id}``."""
        blob, ref = unit.new_ec2_credential(
            user_id=self.user['id'], project_id=self.project_id
        )
        blob['app_cred_id'] = uuid.uuid4().hex
        ref['blob'] = json.dumps(blob)
        r = self.post('/credentials', body={'credential': ref})
        self.assertValidCredentialResponse(r, ref)
        credential_id = r.result.get('credential')['id']
        # Try changing to a different app cred
        blob['app_cred_id'] = uuid.uuid4().hex
        update_ref = {'blob': json.dumps(blob)}
        self.patch(
            f'/credentials/{credential_id}',
            body={'credential': update_ref},
            expected_status=http.client.BAD_REQUEST,
        )
        # Try removing the app cred
        del blob['app_cred_id']
        update_ref = {'blob': json.dumps(blob)}
        self.patch(
            f'/credentials/{credential_id}',
            body={'credential': update_ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_update_ec2_credential_change_access_token_id(self):
        """Call ``PATCH /credentials/{credential_id}``."""
        blob, ref = unit.new_ec2_credential(
            user_id=self.user['id'], project_id=self.project_id
        )
        blob['access_token_id'] = uuid.uuid4().hex
        ref['blob'] = json.dumps(blob)
        r = self.post('/credentials', body={'credential': ref})
        self.assertValidCredentialResponse(r, ref)
        credential_id = r.result.get('credential')['id']
        # Try changing to a different access token
        blob['access_token_id'] = uuid.uuid4().hex
        update_ref = {'blob': json.dumps(blob)}
        self.patch(
            f'/credentials/{credential_id}',
            body={'credential': update_ref},
            expected_status=http.client.BAD_REQUEST,
        )
        # Try removing the access token
        del blob['access_token_id']
        update_ref = {'blob': json.dumps(blob)}
        self.patch(
            f'/credentials/{credential_id}',
            body={'credential': update_ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_update_ec2_credential_change_access_id(self):
        """Call ``PATCH /credentials/{credential_id}``."""
        blob, ref = unit.new_ec2_credential(
            user_id=self.user['id'], project_id=self.project_id
        )
        blob['access_id'] = uuid.uuid4().hex
        ref['blob'] = json.dumps(blob)
        r = self.post('/credentials', body={'credential': ref})
        self.assertValidCredentialResponse(r, ref)
        credential_id = r.result.get('credential')['id']
        # Try changing to a different access_id
        blob['access_id'] = uuid.uuid4().hex
        update_ref = {'blob': json.dumps(blob)}
        self.patch(
            f'/credentials/{credential_id}',
            body={'credential': update_ref},
            expected_status=http.client.BAD_REQUEST,
        )
        # Try removing the access_id
        del blob['access_id']
        update_ref = {'blob': json.dumps(blob)}
        self.patch(
            f'/credentials/{credential_id}',
            body={'credential': update_ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_delete_credential(self):
        """Call ``DELETE /credentials/{credential_id}``."""
        self.delete(
            '/credentials/{credential_id}'.format(
                credential_id=self.credential['id']
            )
        )

    def test_delete_credential_retries_on_deadlock(self):
        patcher = mock.patch(
            'sqlalchemy.orm.query.Query.delete', autospec=True
        )

        class FakeDeadlock:
            def __init__(self, mock_patcher):
                self.deadlock_count = 2
                self.mock_patcher = mock_patcher
                self.patched = True

            def __call__(self, *args, **kwargs):
                if self.deadlock_count > 1:
                    self.deadlock_count -= 1
                else:
                    self.mock_patcher.stop()
                    self.patched = False
                raise oslo_db_exception.DBDeadlock

        sql_delete_mock = patcher.start()
        side_effect = FakeDeadlock(patcher)
        sql_delete_mock.side_effect = side_effect

        try:
            PROVIDERS.credential_api.delete_credentials_for_user(
                user_id=self.user['id']
            )
        finally:
            if side_effect.patched:
                patcher.stop()

        # initial attempt + 1 retry
        self.assertEqual(sql_delete_mock.call_count, 2)

    def test_create_ec2_credential(self):
        """Call ``POST /credentials`` for creating ec2 credential."""
        blob, ref = unit.new_ec2_credential(
            user_id=self.user['id'], project_id=self.project_id
        )
        r = self.post('/credentials', body={'credential': ref})
        self.assertValidCredentialResponse(r, ref)
        # Assert credential id is same as hash of access key id for
        # ec2 credentials
        access = blob['access'].encode('utf-8')
        self.assertEqual(
            hashlib.sha256(access).hexdigest(), r.result['credential']['id']
        )
        # Create second ec2 credential with the same access key id and check
        # for conflict.
        self.post(
            '/credentials',
            body={'credential': ref},
            expected_status=http.client.CONFLICT,
        )

    def test_get_ec2_dict_blob(self):
        """Ensure non-JSON blob data is correctly converted."""
        expected_blob, credential_id = self._create_dict_blob_credential()

        r = self.get(f'/credentials/{credential_id}')

        # use json.loads to transform the blobs back into Python dictionaries
        # to avoid problems with the keys being in different orders.
        self.assertEqual(
            json.loads(expected_blob),
            json.loads(r.result['credential']['blob']),
        )

    def test_list_ec2_dict_blob(self):
        """Ensure non-JSON blob data is correctly converted."""
        expected_blob, credential_id = self._create_dict_blob_credential()

        list_r = self.get('/credentials')
        list_creds = list_r.result['credentials']
        list_ids = [r['id'] for r in list_creds]
        self.assertIn(credential_id, list_ids)
        # use json.loads to transform the blobs back into Python dictionaries
        # to avoid problems with the keys being in different orders.
        for r in list_creds:
            if r['id'] == credential_id:
                self.assertEqual(
                    json.loads(expected_blob), json.loads(r['blob'])
                )

    def test_create_non_ec2_credential(self):
        """Test creating non-ec2 credential.

        Call ``POST /credentials``.
        """
        blob, ref = unit.new_cert_credential(user_id=self.user['id'])

        r = self.post('/credentials', body={'credential': ref})
        self.assertValidCredentialResponse(r, ref)
        # Assert credential id is not same as hash of access key id for
        # non-ec2 credentials
        access = blob['access'].encode('utf-8')
        self.assertNotEqual(
            hashlib.sha256(access).hexdigest(), r.result['credential']['id']
        )

    def test_create_ec2_credential_with_missing_project_id(self):
        """Test Creating ec2 credential with missing project_id.

        Call ``POST /credentials``.
        """
        _, ref = unit.new_ec2_credential(
            user_id=self.user['id'], project_id=None
        )
        # Assert bad request status when missing project_id
        self.post(
            '/credentials',
            body={'credential': ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_create_ec2_credential_with_invalid_blob(self):
        """Test creating ec2 credential with invalid blob.

        Call ``POST /credentials``.
        """
        ref = unit.new_credential_ref(
            user_id=self.user['id'],
            project_id=self.project_id,
            blob='{"abc":"def"d}',
            type=CRED_TYPE_EC2,
        )
        # Assert bad request status when request contains invalid blob
        response = self.post(
            '/credentials',
            body={'credential': ref},
            expected_status=http.client.BAD_REQUEST,
        )
        self.assertValidErrorResponse(response)

    def test_create_credential_with_admin_token(self):
        # Make sure we can create credential with the static admin token
        ref = unit.new_credential_ref(user_id=self.user['id'])
        r = self.post(
            '/credentials',
            body={'credential': ref},
            token=self.get_admin_token(),
        )
        self.assertValidCredentialResponse(r, ref)


class TestCredentialTrustScoped(CredentialBaseTestCase):
    """Test credential with trust scoped token."""

    def setUp(self):
        super().setUp()

        self.trustee_user = unit.new_user_ref(domain_id=self.domain_id)
        password = self.trustee_user['password']
        self.trustee_user = PROVIDERS.identity_api.create_user(
            self.trustee_user
        )
        self.trustee_user['password'] = password
        self.trustee_user_id = self.trustee_user['id']
        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'credential',
                credential_fernet.MAX_ACTIVE_KEYS,
            )
        )

    def config_overrides(self):
        super().config_overrides()
        self.config_fixture.config(group='trust')

    def test_trust_scoped_ec2_credential(self):
        """Test creating trust scoped ec2 credential.

        Call ``POST /credentials``.
        """
        # Create the trust
        ref = unit.new_trust_ref(
            trustor_user_id=self.user_id,
            trustee_user_id=self.trustee_user_id,
            project_id=self.project_id,
            impersonation=True,
            expires={'minutes': 1},
            role_ids=[self.role_id],
        )
        del ref['id']
        r = self.post('/OS-TRUST/trusts', body={'trust': ref})
        trust = self.assertValidTrustResponse(r)

        # Get a trust scoped token
        auth_data = self.build_authentication_request(
            user_id=self.trustee_user['id'],
            password=self.trustee_user['password'],
            trust_id=trust['id'],
        )
        r = self.v3_create_token(auth_data)
        self.assertValidProjectScopedTokenResponse(r, self.user)
        trust_id = r.result['token']['OS-TRUST:trust']['id']
        token_id = r.headers.get('X-Subject-Token')

        # Create the credential with the trust scoped token
        blob, ref = unit.new_ec2_credential(
            user_id=self.user_id, project_id=self.project_id
        )
        r = self.post('/credentials', body={'credential': ref}, token=token_id)

        # We expect the response blob to contain the trust_id
        ret_ref = ref.copy()
        ret_blob = blob.copy()
        ret_blob['trust_id'] = trust_id
        ret_ref['blob'] = json.dumps(ret_blob)
        self.assertValidCredentialResponse(r, ref=ret_ref)

        # Assert credential id is same as hash of access key id for
        # ec2 credentials
        access = blob['access'].encode('utf-8')
        self.assertEqual(
            hashlib.sha256(access).hexdigest(), r.result['credential']['id']
        )

        # Create a role assignment to ensure that it is ignored and only the
        # trust-delegated roles are used
        role = unit.new_role_ref(name='reader')
        role_id = role['id']
        PROVIDERS.role_api.create_role(role_id, role)
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            self.user_id, self.project_id, role_id
        )

        ret_blob = json.loads(r.result['credential']['blob'])
        ec2token = self._test_get_token(
            access=ret_blob['access'], secret=ret_blob['secret']
        )
        ec2_roles = [role['id'] for role in ec2token['roles']]
        self.assertIn(self.role_id, ec2_roles)
        self.assertNotIn(role_id, ec2_roles)

        # Create second ec2 credential with the same access key id and check
        # for conflict.
        self.post(
            '/credentials',
            body={'credential': ref},
            token=token_id,
            expected_status=http.client.CONFLICT,
        )


class TestCredentialAppCreds(CredentialBaseTestCase):
    """Test credential with application credential token."""

    def setUp(self):
        super().setUp()
        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'credential',
                credential_fernet.MAX_ACTIVE_KEYS,
            )
        )

    def test_app_cred_ec2_credential(self):
        """Test creating ec2 credential from an application credential.

        Call ``POST /credentials``.
        """
        # Create the app cred
        ref = unit.new_application_credential_ref(roles=[{'id': self.role_id}])
        del ref['id']
        r = self.post(
            f'/users/{self.user_id}/application_credentials',
            body={'application_credential': ref},
        )
        app_cred = r.result['application_credential']

        # Get an application credential token
        auth_data = self.build_authentication_request(
            app_cred_id=app_cred['id'], secret=app_cred['secret']
        )
        r = self.v3_create_token(auth_data)
        token_id = r.headers.get('X-Subject-Token')

        # Create the credential with the app cred token
        blob, ref = unit.new_ec2_credential(
            user_id=self.user_id, project_id=self.project_id
        )
        r = self.post('/credentials', body={'credential': ref}, token=token_id)

        # We expect the response blob to contain the app_cred_id
        ret_ref = ref.copy()
        ret_blob = blob.copy()
        ret_blob['app_cred_id'] = app_cred['id']
        ret_ref['blob'] = json.dumps(ret_blob)
        self.assertValidCredentialResponse(r, ref=ret_ref)

        # Assert credential id is same as hash of access key id for
        # ec2 credentials
        access = blob['access'].encode('utf-8')
        self.assertEqual(
            hashlib.sha256(access).hexdigest(), r.result['credential']['id']
        )

        # Create a role assignment to ensure that it is ignored and only the
        # roles in the app cred are used
        role = unit.new_role_ref(name='reader')
        role_id = role['id']
        PROVIDERS.role_api.create_role(role_id, role)
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            self.user_id, self.project_id, role_id
        )

        ret_blob = json.loads(r.result['credential']['blob'])
        ec2token = self._test_get_token(
            access=ret_blob['access'], secret=ret_blob['secret']
        )
        ec2_roles = [role['id'] for role in ec2token['roles']]
        self.assertIn(self.role_id, ec2_roles)
        self.assertNotIn(role_id, ec2_roles)

        # Create second ec2 credential with the same access key id and check
        # for conflict.
        self.post(
            '/credentials',
            body={'credential': ref},
            token=token_id,
            expected_status=http.client.CONFLICT,
        )


class TestCredentialAccessToken(CredentialBaseTestCase):
    """Test credential with access token."""

    def setUp(self):
        super().setUp()
        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'credential',
                credential_fernet.MAX_ACTIVE_KEYS,
            )
        )
        self.base_url = 'http://localhost/v3'

    def _urllib_parse_qs_text_keys(self, content):
        results = urllib.parse.parse_qs(content)
        return {key.decode('utf-8'): value for key, value in results.items()}

    def _create_single_consumer(self):
        endpoint = '/OS-OAUTH1/consumers'

        ref = {'description': uuid.uuid4().hex}
        resp = self.post(endpoint, body={'consumer': ref})
        return resp.result['consumer']

    def _create_request_token(self, consumer, project_id, base_url=None):
        endpoint = '/OS-OAUTH1/request_token'
        client = oauth1.Client(
            consumer['key'],
            client_secret=consumer['secret'],
            signature_method=oauth1.SIG_HMAC,
            callback_uri="oob",
        )
        headers = {'requested_project_id': project_id}
        if not base_url:
            base_url = self.base_url
        url, headers, body = client.sign(
            base_url + endpoint, http_method='POST', headers=headers
        )
        return endpoint, headers

    def _create_access_token(self, consumer, token, base_url=None):
        endpoint = '/OS-OAUTH1/access_token'
        client = oauth1.Client(
            consumer['key'],
            client_secret=consumer['secret'],
            resource_owner_key=token.key,
            resource_owner_secret=token.secret,
            signature_method=oauth1.SIG_HMAC,
            verifier=token.verifier,
        )
        if not base_url:
            base_url = self.base_url
        url, headers, body = client.sign(
            base_url + endpoint, http_method='POST'
        )
        headers.update({'Content-Type': 'application/json'})
        return endpoint, headers

    def _get_oauth_token(self, consumer, token):
        client = oauth1.Client(
            consumer['key'],
            client_secret=consumer['secret'],
            resource_owner_key=token.key,
            resource_owner_secret=token.secret,
            signature_method=oauth1.SIG_HMAC,
        )
        endpoint = '/auth/tokens'
        url, headers, body = client.sign(
            self.base_url + endpoint, http_method='POST'
        )
        headers.update({'Content-Type': 'application/json'})
        ref = {'auth': {'identity': {'oauth1': {}, 'methods': ['oauth1']}}}
        return endpoint, headers, ref

    def _authorize_request_token(self, request_id):
        if isinstance(request_id, bytes):
            request_id = request_id.decode()
        return f'/OS-OAUTH1/authorize/{request_id}'

    def _get_access_token(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}

        url, headers = self._create_request_token(consumer, self.project_id)
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = self._urllib_parse_qs_text_keys(content.result)
        request_key = credentials['oauth_token'][0]
        request_secret = credentials['oauth_token_secret'][0]
        request_token = oauth1.Token(request_key, request_secret)

        url = self._authorize_request_token(request_key)
        body = {'roles': [{'id': self.role_id}]}
        resp = self.put(url, body=body, expected_status=http.client.OK)
        verifier = resp.result['token']['oauth_verifier']

        request_token.set_verifier(verifier)
        url, headers = self._create_access_token(consumer, request_token)
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = self._urllib_parse_qs_text_keys(content.result)
        access_key = credentials['oauth_token'][0]
        access_secret = credentials['oauth_token_secret'][0]
        access_token = oauth1.Token(access_key, access_secret)

        url, headers, body = self._get_oauth_token(consumer, access_token)
        content = self.post(url, headers=headers, body=body)
        return access_key, content.headers['X-Subject-Token']

    def test_access_token_ec2_credential(self):
        """Test creating ec2 credential from an oauth access token.

        Call ``POST /credentials``.
        """
        access_key, token_id = self._get_access_token()

        # Create the credential with the access token
        blob, ref = unit.new_ec2_credential(
            user_id=self.user_id, project_id=self.project_id
        )
        r = self.post('/credentials', body={'credential': ref}, token=token_id)

        # We expect the response blob to contain the access_token_id
        ret_ref = ref.copy()
        ret_blob = blob.copy()
        ret_blob['access_token_id'] = access_key.decode('utf-8')
        ret_ref['blob'] = json.dumps(ret_blob)
        self.assertValidCredentialResponse(r, ref=ret_ref)

        # Assert credential id is same as hash of access key id for
        # ec2 credentials
        access = blob['access'].encode('utf-8')
        self.assertEqual(
            hashlib.sha256(access).hexdigest(), r.result['credential']['id']
        )

        # Create a role assignment to ensure that it is ignored and only the
        # roles in the access token are used
        role = unit.new_role_ref(name='reader')
        role_id = role['id']
        PROVIDERS.role_api.create_role(role_id, role)
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            self.user_id, self.project_id, role_id
        )

        ret_blob = json.loads(r.result['credential']['blob'])
        ec2token = self._test_get_token(
            access=ret_blob['access'], secret=ret_blob['secret']
        )
        ec2_roles = [role['id'] for role in ec2token['roles']]
        self.assertIn(self.role_id, ec2_roles)
        self.assertNotIn(role_id, ec2_roles)


class TestCredentialEc2(CredentialBaseTestCase):
    """Test v3 credential compatibility with ec2tokens."""

    def test_ec2_credential_signature_validate(self):
        """Test signature validation with a v3 ec2 credential."""
        blob, ref = unit.new_ec2_credential(
            user_id=self.user['id'], project_id=self.project_id
        )
        r = self.post('/credentials', body={'credential': ref})
        self.assertValidCredentialResponse(r, ref)
        # Assert credential id is same as hash of access key id
        access = blob['access'].encode('utf-8')
        self.assertEqual(
            hashlib.sha256(access).hexdigest(), r.result['credential']['id']
        )

        cred_blob = json.loads(r.result['credential']['blob'])
        self.assertEqual(blob, cred_blob)
        self._test_get_token(
            access=cred_blob['access'], secret=cred_blob['secret']
        )

    def test_ec2_credential_signature_validate_legacy(self):
        """Test signature validation with a legacy v3 ec2 credential."""
        cred_json, _ = self._create_dict_blob_credential()
        cred_blob = json.loads(cred_json)
        self._test_get_token(
            access=cred_blob['access'], secret=cred_blob['secret']
        )

    def _get_ec2_cred_uri(self):
        return f'/users/{self.user_id}/credentials/OS-EC2'

    def _get_ec2_cred(self):
        uri = self._get_ec2_cred_uri()
        r = self.post(uri, body={'tenant_id': self.project_id})
        return r.result['credential']

    def test_ec2_create_credential(self):
        """Test ec2 credential creation."""
        ec2_cred = self._get_ec2_cred()
        self.assertEqual(self.user_id, ec2_cred['user_id'])
        self.assertEqual(self.project_id, ec2_cred['tenant_id'])
        self.assertIsNone(ec2_cred['trust_id'])
        self._test_get_token(
            access=ec2_cred['access'], secret=ec2_cred['secret']
        )
        uri = '/'.join([self._get_ec2_cred_uri(), ec2_cred['access']])
        self.assertThat(ec2_cred['links']['self'], matchers.EndsWith(uri))

    def test_ec2_get_credential(self):
        ec2_cred = self._get_ec2_cred()
        uri = '/'.join([self._get_ec2_cred_uri(), ec2_cred['access']])
        r = self.get(uri)
        self.assertDictEqual(ec2_cred, r.result['credential'])
        self.assertThat(ec2_cred['links']['self'], matchers.EndsWith(uri))

    def test_ec2_cannot_get_non_ec2_credential(self):
        access_key = uuid.uuid4().hex
        cred_id = utils.hash_access_key(access_key)
        non_ec2_cred = unit.new_credential_ref(
            user_id=self.user_id, project_id=self.project_id
        )
        non_ec2_cred['id'] = cred_id
        PROVIDERS.credential_api.create_credential(cred_id, non_ec2_cred)
        uri = '/'.join([self._get_ec2_cred_uri(), access_key])
        # if access_key is not found, ec2 controller raises Unauthorized
        # exception
        self.get(uri, expected_status=http.client.UNAUTHORIZED)

    def test_ec2_list_credentials(self):
        """Test ec2 credential listing."""
        self._get_ec2_cred()
        uri = self._get_ec2_cred_uri()
        r = self.get(uri)
        cred_list = r.result['credentials']
        self.assertEqual(1, len(cred_list))
        self.assertThat(r.result['links']['self'], matchers.EndsWith(uri))

        # non-EC2 credentials won't be fetched
        non_ec2_cred = unit.new_credential_ref(
            user_id=self.user_id, project_id=self.project_id
        )
        non_ec2_cred['type'] = uuid.uuid4().hex
        PROVIDERS.credential_api.create_credential(
            non_ec2_cred['id'], non_ec2_cred
        )
        r = self.get(uri)
        cred_list_2 = r.result['credentials']
        # still one element because non-EC2 credentials are not returned.
        self.assertEqual(1, len(cred_list_2))
        self.assertEqual(cred_list[0], cred_list_2[0])

    def test_ec2_delete_credential(self):
        """Test ec2 credential deletion."""
        ec2_cred = self._get_ec2_cred()
        uri = '/'.join([self._get_ec2_cred_uri(), ec2_cred['access']])
        cred_from_credential_api = (
            PROVIDERS.credential_api.list_credentials_for_user(
                self.user_id, type=CRED_TYPE_EC2
            )
        )
        self.assertEqual(1, len(cred_from_credential_api))
        self.delete(uri)
        self.assertRaises(
            exception.CredentialNotFound,
            PROVIDERS.credential_api.get_credential,
            cred_from_credential_api[0]['id'],
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from unittest import mock
import uuid

from oslo_config import cfg
from testtools import matchers

from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone.tests import unit
from keystone.tests.unit import default_fixtures
from keystone.tests.unit.ksfixtures import database

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class TestResourceManagerNoFixtures(unit.SQLDriverOverrides, unit.TestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(database.Database())
        self.load_backends()
        PROVIDERS.resource_api.create_domain(
            default_fixtures.ROOT_DOMAIN['id'], default_fixtures.ROOT_DOMAIN
        )

    def test_update_project_name_conflict(self):
        name = uuid.uuid4().hex
        description = uuid.uuid4().hex
        domain_attrs = {
            'id': CONF.identity.default_domain_id,
            'name': name,
            'description': description,
        }
        domain = PROVIDERS.resource_api.create_domain(
            CONF.identity.default_domain_id, domain_attrs
        )
        project1 = unit.new_project_ref(
            domain_id=domain['id'], name=uuid.uuid4().hex
        )
        PROVIDERS.resource_api.create_project(project1['id'], project1)
        project2 = unit.new_project_ref(
            domain_id=domain['id'], name=uuid.uuid4().hex
        )
        project = PROVIDERS.resource_api.create_project(
            project2['id'], project2
        )

        self.assertRaises(
            exception.Conflict,
            PROVIDERS.resource_api.update_project,
            project['id'],
            {'name': project1['name'], 'id': project['id']},
        )


class DomainConfigDriverTests:
    def _domain_config_crud(self, sensitive):
        domain = uuid.uuid4().hex
        group = uuid.uuid4().hex
        option = uuid.uuid4().hex
        value = uuid.uuid4().hex
        config = {
            'group': group,
            'option': option,
            'value': value,
            'sensitive': sensitive,
        }
        self.driver.create_config_options(domain, [config])
        res = self.driver.get_config_option(domain, group, option, sensitive)
        config.pop('sensitive')
        self.assertEqual(config, res)

        value = uuid.uuid4().hex
        config = {
            'group': group,
            'option': option,
            'value': value,
            'sensitive': sensitive,
        }
        self.driver.update_config_options(domain, [config])
        res = self.driver.get_config_option(domain, group, option, sensitive)
        config.pop('sensitive')
        self.assertEqual(config, res)

        self.driver.delete_config_options(domain, group, option)
        self.assertRaises(
            exception.DomainConfigNotFound,
            self.driver.get_config_option,
            domain,
            group,
            option,
            sensitive,
        )
        # ...and silent if we try to delete it again
        self.driver.delete_config_options(domain, group, option)

    def test_whitelisted_domain_config_crud(self):
        self._domain_config_crud(sensitive=False)

    def test_sensitive_domain_config_crud(self):
        self._domain_config_crud(sensitive=True)

    def _list_domain_config(self, sensitive):
        """Test listing by combination of domain, group & option."""
        config1 = {
            'group': uuid.uuid4().hex,
            'option': uuid.uuid4().hex,
            'value': uuid.uuid4().hex,
            'sensitive': sensitive,
        }
        # Put config2 in the same group as config1
        config2 = {
            'group': config1['group'],
            'option': uuid.uuid4().hex,
            'value': uuid.uuid4().hex,
            'sensitive': sensitive,
        }
        config3 = {
            'group': uuid.uuid4().hex,
            'option': uuid.uuid4().hex,
            'value': 100,
            'sensitive': sensitive,
        }
        domain = uuid.uuid4().hex

        self.driver.create_config_options(domain, [config1, config2, config3])

        for config in [config1, config2, config3]:
            config.pop('sensitive')
        # Try listing all items from a domain
        res = self.driver.list_config_options(domain, sensitive=sensitive)
        self.assertThat(res, matchers.HasLength(3))
        for res_entry in res:
            self.assertIn(res_entry, [config1, config2, config3])

        # Try listing by domain and group
        res = self.driver.list_config_options(
            domain, group=config1['group'], sensitive=sensitive
        )
        self.assertThat(res, matchers.HasLength(2))
        for res_entry in res:
            self.assertIn(res_entry, [config1, config2])

        # Try listing by domain, group and option
        res = self.driver.list_config_options(
            domain,
            group=config2['group'],
            option=config2['option'],
            sensitive=sensitive,
        )
        self.assertThat(res, matchers.HasLength(1))
        self.assertEqual(config2, res[0])

    def test_list_whitelisted_domain_config_crud(self):
        self._list_domain_config(False)

    def test_list_sensitive_domain_config_crud(self):
        self._list_domain_config(True)

    def _delete_domain_configs(self, sensitive):
        """Test deleting by combination of domain, group & option."""
        config1 = {
            'group': uuid.uuid4().hex,
            'option': uuid.uuid4().hex,
            'value': uuid.uuid4().hex,
            'sensitive': sensitive,
        }
        # Put config2 and config3 in the same group as config1
        config2 = {
            'group': config1['group'],
            'option': uuid.uuid4().hex,
            'value': uuid.uuid4().hex,
            'sensitive': sensitive,
        }
        config3 = {
            'group': config1['group'],
            'option': uuid.uuid4().hex,
            'value': uuid.uuid4().hex,
            'sensitive': sensitive,
        }
        config4 = {
            'group': uuid.uuid4().hex,
            'option': uuid.uuid4().hex,
            'value': uuid.uuid4().hex,
            'sensitive': sensitive,
        }
        domain = uuid.uuid4().hex

        self.driver.create_config_options(
            domain, [config1, config2, config3, config4]
        )

        for config in [config1, config2, config3, config4]:
            config.pop('sensitive')
        # Try deleting by domain, group and option
        res = self.driver.delete_config_options(
            domain, group=config2['group'], option=config2['option']
        )
        res = self.driver.list_config_options(domain, sensitive=sensitive)
        self.assertThat(res, matchers.HasLength(3))
        for res_entry in res:
            self.assertIn(res_entry, [config1, config3, config4])

        # Try deleting by domain and group
        res = self.driver.delete_config_options(domain, group=config4['group'])
        res = self.driver.list_config_options(domain, sensitive=sensitive)
        self.assertThat(res, matchers.HasLength(2))
        for res_entry in res:
            self.assertIn(res_entry, [config1, config3])

        # Try deleting all items from a domain
        res = self.driver.delete_config_options(domain)
        res = self.driver.list_config_options(domain, sensitive=sensitive)
        self.assertThat(res, matchers.HasLength(0))

    def test_delete_whitelisted_domain_configs(self):
        self._delete_domain_configs(False)

    def test_delete_sensitive_domain_configs(self):
        self._delete_domain_configs(True)

    def _create_domain_config_twice(self, sensitive):
        """Test create the same option twice just overwrites."""
        config = {
            'group': uuid.uuid4().hex,
            'option': uuid.uuid4().hex,
            'value': uuid.uuid4().hex,
            'sensitive': sensitive,
        }
        domain = uuid.uuid4().hex

        self.driver.create_config_options(domain, [config])
        config['value'] = uuid.uuid4().hex
        self.driver.create_config_options(domain, [config])
        res = self.driver.get_config_option(
            domain, config['group'], config['option'], sensitive
        )
        config.pop('sensitive')
        self.assertEqual(config, res)

    def test_create_whitelisted_domain_config_twice(self):
        self._create_domain_config_twice(False)

    def test_create_sensitive_domain_config_twice(self):
        self._create_domain_config_twice(True)


class DomainConfigTests:
    def setUp(self):
        self.domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(self.domain['id'], self.domain)
        self.addCleanup(self.clean_up_domain)

    def clean_up_domain(self):
        # NOTE(henry-nash): Deleting the domain will also delete any domain
        # configs for this domain.
        self.domain['enabled'] = False
        PROVIDERS.resource_api.update_domain(self.domain['id'], self.domain)
        PROVIDERS.resource_api.delete_domain(self.domain['id'])
        del self.domain

    def test_create_domain_config_including_sensitive_option(self):
        config = {
            'ldap': {
                'url': uuid.uuid4().hex,
                'user_tree_dn': uuid.uuid4().hex,
                'password': uuid.uuid4().hex,
            }
        }
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)

        # password is sensitive, so check that the whitelisted portion and
        # the sensitive piece have been stored in the appropriate locations.
        res = PROVIDERS.domain_config_api.get_config(self.domain['id'])
        config_whitelisted = copy.deepcopy(config)
        config_whitelisted['ldap'].pop('password')
        self.assertEqual(config_whitelisted, res)
        res = PROVIDERS.domain_config_api.driver.get_config_option(
            self.domain['id'], 'ldap', 'password', sensitive=True
        )
        self.assertEqual(config['ldap']['password'], res['value'])

        # Finally, use the non-public API to get back the whole config
        res = PROVIDERS.domain_config_api.get_config_with_sensitive_info(
            self.domain['id']
        )
        self.assertEqual(config, res)

    def test_get_partial_domain_config(self):
        config = {
            'ldap': {
                'url': uuid.uuid4().hex,
                'user_tree_dn': uuid.uuid4().hex,
                'password': uuid.uuid4().hex,
            },
            'identity': {'driver': uuid.uuid4().hex},
        }
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)

        res = PROVIDERS.domain_config_api.get_config(
            self.domain['id'], group='identity'
        )
        config_partial = copy.deepcopy(config)
        config_partial.pop('ldap')
        self.assertEqual(config_partial, res)
        res = PROVIDERS.domain_config_api.get_config(
            self.domain['id'], group='ldap', option='user_tree_dn'
        )
        self.assertEqual({'user_tree_dn': config['ldap']['user_tree_dn']}, res)
        # ...but we should fail to get a sensitive option
        self.assertRaises(
            exception.DomainConfigNotFound,
            PROVIDERS.domain_config_api.get_config,
            self.domain['id'],
            group='ldap',
            option='password',
        )

    def test_delete_partial_domain_config(self):
        config = {
            'ldap': {
                'url': uuid.uuid4().hex,
                'user_tree_dn': uuid.uuid4().hex,
                'password': uuid.uuid4().hex,
            },
            'identity': {'driver': uuid.uuid4().hex},
        }
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)

        PROVIDERS.domain_config_api.delete_config(
            self.domain['id'], group='identity'
        )
        config_partial = copy.deepcopy(config)
        config_partial.pop('identity')
        config_partial['ldap'].pop('password')
        res = PROVIDERS.domain_config_api.get_config(self.domain['id'])
        self.assertEqual(config_partial, res)

        PROVIDERS.domain_config_api.delete_config(
            self.domain['id'], group='ldap', option='url'
        )
        config_partial = copy.deepcopy(config_partial)
        config_partial['ldap'].pop('url')
        res = PROVIDERS.domain_config_api.get_config(self.domain['id'])
        self.assertEqual(config_partial, res)

    def test_get_options_not_in_domain_config(self):
        self.assertRaises(
            exception.DomainConfigNotFound,
            PROVIDERS.domain_config_api.get_config,
            self.domain['id'],
        )
        config = {'ldap': {'url': uuid.uuid4().hex}}

        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)

        self.assertRaises(
            exception.DomainConfigNotFound,
            PROVIDERS.domain_config_api.get_config,
            self.domain['id'],
            group='identity',
        )
        self.assertRaises(
            exception.DomainConfigNotFound,
            PROVIDERS.domain_config_api.get_config,
            self.domain['id'],
            group='ldap',
            option='user_tree_dn',
        )

    def test_get_sensitive_config(self):
        config = {
            'ldap': {
                'url': uuid.uuid4().hex,
                'user_tree_dn': uuid.uuid4().hex,
                'password': uuid.uuid4().hex,
            },
            'identity': {'driver': uuid.uuid4().hex},
        }
        res = PROVIDERS.domain_config_api.get_config_with_sensitive_info(
            self.domain['id']
        )
        self.assertEqual({}, res)
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)
        res = PROVIDERS.domain_config_api.get_config_with_sensitive_info(
            self.domain['id']
        )
        self.assertEqual(config, res)

    def test_update_partial_domain_config(self):
        config = {
            'ldap': {
                'url': uuid.uuid4().hex,
                'user_tree_dn': uuid.uuid4().hex,
                'password': uuid.uuid4().hex,
            },
            'identity': {'driver': uuid.uuid4().hex},
        }
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)

        # Try updating a group
        new_config = {
            'ldap': {'url': uuid.uuid4().hex, 'user_filter': uuid.uuid4().hex}
        }
        res = PROVIDERS.domain_config_api.update_config(
            self.domain['id'], new_config, group='ldap'
        )
        expected_config = copy.deepcopy(config)
        expected_config['ldap']['url'] = new_config['ldap']['url']
        expected_config['ldap']['user_filter'] = new_config['ldap'][
            'user_filter'
        ]
        expected_full_config = copy.deepcopy(expected_config)
        expected_config['ldap'].pop('password')
        res = PROVIDERS.domain_config_api.get_config(self.domain['id'])
        self.assertEqual(expected_config, res)
        # The sensitive option should still exist
        res = PROVIDERS.domain_config_api.get_config_with_sensitive_info(
            self.domain['id']
        )
        self.assertEqual(expected_full_config, res)

        # Try updating a single whitelisted option
        PROVIDERS.domain_config_api.delete_config(self.domain['id'])
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)
        new_config = {'url': uuid.uuid4().hex}
        res = PROVIDERS.domain_config_api.update_config(
            self.domain['id'], new_config, group='ldap', option='url'
        )

        # Make sure whitelisted and full config is updated
        expected_whitelisted_config = copy.deepcopy(config)
        expected_whitelisted_config['ldap']['url'] = new_config['url']
        expected_full_config = copy.deepcopy(expected_whitelisted_config)
        expected_whitelisted_config['ldap'].pop('password')
        self.assertEqual(expected_whitelisted_config, res)
        res = PROVIDERS.domain_config_api.get_config(self.domain['id'])
        self.assertEqual(expected_whitelisted_config, res)
        res = PROVIDERS.domain_config_api.get_config_with_sensitive_info(
            self.domain['id']
        )
        self.assertEqual(expected_full_config, res)

        # Try updating a single sensitive option
        PROVIDERS.domain_config_api.delete_config(self.domain['id'])
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)
        new_config = {'password': uuid.uuid4().hex}
        res = PROVIDERS.domain_config_api.update_config(
            self.domain['id'], new_config, group='ldap', option='password'
        )
        # The whitelisted config should not have changed...
        expected_whitelisted_config = copy.deepcopy(config)
        expected_full_config = copy.deepcopy(config)
        expected_whitelisted_config['ldap'].pop('password')
        self.assertEqual(expected_whitelisted_config, res)
        res = PROVIDERS.domain_config_api.get_config(self.domain['id'])
        self.assertEqual(expected_whitelisted_config, res)
        expected_full_config['ldap']['password'] = new_config['password']
        res = PROVIDERS.domain_config_api.get_config_with_sensitive_info(
            self.domain['id']
        )
        # ...but the sensitive piece should have.
        self.assertEqual(expected_full_config, res)

    def test_update_invalid_partial_domain_config(self):
        config = {
            'ldap': {
                'url': uuid.uuid4().hex,
                'user_tree_dn': uuid.uuid4().hex,
                'password': uuid.uuid4().hex,
            },
            'identity': {'driver': uuid.uuid4().hex},
        }
        # An extra group, when specifying one group should fail
        self.assertRaises(
            exception.InvalidDomainConfig,
            PROVIDERS.domain_config_api.update_config,
            self.domain['id'],
            config,
            group='ldap',
        )
        # An extra option, when specifying one option should fail
        self.assertRaises(
            exception.InvalidDomainConfig,
            PROVIDERS.domain_config_api.update_config,
            self.domain['id'],
            config['ldap'],
            group='ldap',
            option='url',
        )

        # Now try the right number of groups/options, but just not
        # ones that are in the config provided
        config = {'ldap': {'user_tree_dn': uuid.uuid4().hex}}
        self.assertRaises(
            exception.InvalidDomainConfig,
            PROVIDERS.domain_config_api.update_config,
            self.domain['id'],
            config,
            group='identity',
        )
        self.assertRaises(
            exception.InvalidDomainConfig,
            PROVIDERS.domain_config_api.update_config,
            self.domain['id'],
            config['ldap'],
            group='ldap',
            option='url',
        )

        # Now some valid groups/options, but just not ones that are in the
        # existing config
        config = {'ldap': {'user_tree_dn': uuid.uuid4().hex}}
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)
        config_wrong_group = {'identity': {'driver': uuid.uuid4().hex}}
        self.assertRaises(
            exception.DomainConfigNotFound,
            PROVIDERS.domain_config_api.update_config,
            self.domain['id'],
            config_wrong_group,
            group='identity',
        )
        config_wrong_option = {'url': uuid.uuid4().hex}
        self.assertRaises(
            exception.DomainConfigNotFound,
            PROVIDERS.domain_config_api.update_config,
            self.domain['id'],
            config_wrong_option,
            group='ldap',
            option='url',
        )

        # And finally just some bad groups/options
        bad_group = uuid.uuid4().hex
        config = {bad_group: {'user': uuid.uuid4().hex}}
        self.assertRaises(
            exception.InvalidDomainConfig,
            PROVIDERS.domain_config_api.update_config,
            self.domain['id'],
            config,
            group=bad_group,
            option='user',
        )
        bad_option = uuid.uuid4().hex
        config = {'ldap': {bad_option: uuid.uuid4().hex}}
        self.assertRaises(
            exception.InvalidDomainConfig,
            PROVIDERS.domain_config_api.update_config,
            self.domain['id'],
            config,
            group='ldap',
            option=bad_option,
        )

    def test_create_invalid_domain_config(self):
        self.assertRaises(
            exception.InvalidDomainConfig,
            PROVIDERS.domain_config_api.create_config,
            self.domain['id'],
            {},
        )
        config = {uuid.uuid4().hex: uuid.uuid4().hex}
        self.assertRaises(
            exception.InvalidDomainConfig,
            PROVIDERS.domain_config_api.create_config,
            self.domain['id'],
            config,
        )
        config = {uuid.uuid4().hex: {uuid.uuid4().hex: uuid.uuid4().hex}}
        self.assertRaises(
            exception.InvalidDomainConfig,
            PROVIDERS.domain_config_api.create_config,
            self.domain['id'],
            config,
        )
        config = {'ldap': {uuid.uuid4().hex: uuid.uuid4().hex}}
        self.assertRaises(
            exception.InvalidDomainConfig,
            PROVIDERS.domain_config_api.create_config,
            self.domain['id'],
            config,
        )
        # Try an option that IS in the standard conf, but neither whitelisted
        # or marked as sensitive
        config = {'identity': {'user_tree_dn': uuid.uuid4().hex}}
        self.assertRaises(
            exception.InvalidDomainConfig,
            PROVIDERS.domain_config_api.create_config,
            self.domain['id'],
            config,
        )

    def test_delete_invalid_partial_domain_config(self):
        config = {'ldap': {'url': uuid.uuid4().hex}}
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)
        # Try deleting a group not in the config
        self.assertRaises(
            exception.DomainConfigNotFound,
            PROVIDERS.domain_config_api.delete_config,
            self.domain['id'],
            group='identity',
        )
        # Try deleting an option not in the config
        self.assertRaises(
            exception.DomainConfigNotFound,
            PROVIDERS.domain_config_api.delete_config,
            self.domain['id'],
            group='ldap',
            option='user_tree_dn',
        )

    def test_sensitive_substitution_in_domain_config(self):
        # Create a config that contains a whitelisted option that requires
        # substitution of a sensitive option.
        config = {
            'ldap': {
                'url': 'my_url/%(password)s',
                'user_tree_dn': uuid.uuid4().hex,
                'password': uuid.uuid4().hex,
            },
            'identity': {'driver': uuid.uuid4().hex},
        }
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)

        # Read back the config with the internal method and ensure that the
        # substitution has taken place.
        res = PROVIDERS.domain_config_api.get_config_with_sensitive_info(
            self.domain['id']
        )
        expected_url = config['ldap']['url'] % {
            'password': config['ldap']['password']
        }
        self.assertEqual(expected_url, res['ldap']['url'])

    def test_invalid_sensitive_substitution_in_domain_config(self):
        """Check that invalid substitutions raise warnings."""
        mock_log = mock.Mock()

        invalid_option_config = {
            'ldap': {
                'user_tree_dn': uuid.uuid4().hex,
                'password': uuid.uuid4().hex,
            },
            'identity': {'driver': uuid.uuid4().hex},
        }

        for invalid_option in [
            'my_url/%(passssword)s',
            'my_url/%(password',
            'my_url/%(password)',
            'my_url/%(password)d',
        ]:
            invalid_option_config['ldap']['url'] = invalid_option
            PROVIDERS.domain_config_api.create_config(
                self.domain['id'], invalid_option_config
            )

            with mock.patch('keystone.resource.core.LOG', mock_log):
                res = (
                    PROVIDERS.domain_config_api.get_config_with_sensitive_info(
                        self.domain['id']
                    )
                )
            mock_log.warning.assert_any_call(mock.ANY, mock.ANY)
            self.assertEqual(
                invalid_option_config['ldap']['url'], res['ldap']['url']
            )

    def test_escaped_sequence_in_domain_config(self):
        """Check that escaped '%(' doesn't get interpreted."""
        mock_log = mock.Mock()

        escaped_option_config = {
            'ldap': {
                'url': 'my_url/%%(password)s',
                'user_tree_dn': uuid.uuid4().hex,
                'password': uuid.uuid4().hex,
            },
            'identity': {'driver': uuid.uuid4().hex},
        }

        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], escaped_option_config
        )

        with mock.patch('keystone.resource.core.LOG', mock_log):
            res = PROVIDERS.domain_config_api.get_config_with_sensitive_info(
                self.domain['id']
            )
        self.assertFalse(mock_log.warn.called)
        # The escaping '%' should have been removed
        self.assertEqual('my_url/%(password)s', res['ldap']['url'])

    @unit.skip_if_cache_disabled('domain_config')
    def test_cache_layer_get_sensitive_config(self):
        config = {
            'ldap': {
                'url': uuid.uuid4().hex,
                'user_tree_dn': uuid.uuid4().hex,
                'password': uuid.uuid4().hex,
            },
            'identity': {'driver': uuid.uuid4().hex},
        }
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)
        # cache the result
        res = PROVIDERS.domain_config_api.get_config_with_sensitive_info(
            self.domain['id']
        )
        self.assertEqual(config, res)

        # delete, bypassing domain config manager api
        PROVIDERS.domain_config_api.delete_config_options(self.domain['id'])

        self.assertDictEqual(
            res,
            PROVIDERS.domain_config_api.get_config_with_sensitive_info(
                self.domain['id']
            ),
        )
        PROVIDERS.domain_config_api.get_config_with_sensitive_info.invalidate(
            PROVIDERS.domain_config_api, self.domain['id']
        )
        self.assertDictEqual(
            {},
            PROVIDERS.domain_config_api.get_config_with_sensitive_info(
                self.domain['id']
            ),
        )

    def test_delete_domain_deletes_configs(self):
        """Test domain deletion clears the domain configs."""
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        config = {
            'ldap': {
                'url': uuid.uuid4().hex,
                'user_tree_dn': uuid.uuid4().hex,
                'password': uuid.uuid4().hex,
            }
        }
        PROVIDERS.domain_config_api.create_config(domain['id'], config)

        # Now delete the domain
        domain['enabled'] = False
        PROVIDERS.resource_api.update_domain(domain['id'], domain)
        PROVIDERS.resource_api.delete_domain(domain['id'])

        # Check domain configs have also been deleted
        self.assertRaises(
            exception.DomainConfigNotFound,
            PROVIDERS.domain_config_api.get_config,
            domain['id'],
        )

        # The get_config_with_sensitive_info does not throw an exception if
        # the config is empty, it just returns an empty dict
        self.assertDictEqual(
            {},
            PROVIDERS.domain_config_api.get_config_with_sensitive_info(
                domain['id']
            ),
        )

    def test_config_registration(self):
        type = uuid.uuid4().hex
        PROVIDERS.domain_config_api.obtain_registration(
            self.domain['id'], type
        )
        PROVIDERS.domain_config_api.release_registration(
            self.domain['id'], type=type
        )

        # Make sure that once someone has it, nobody else can get it.
        # This includes the domain who already has it.
        PROVIDERS.domain_config_api.obtain_registration(
            self.domain['id'], type
        )
        self.assertFalse(
            PROVIDERS.domain_config_api.obtain_registration(
                self.domain['id'], type
            )
        )

        # Make sure we can read who does have it
        self.assertEqual(
            self.domain['id'],
            PROVIDERS.domain_config_api.read_registration(type),
        )

        # Make sure releasing it is silent if the domain specified doesn't
        # have the registration
        domain2 = {'id': uuid.uuid4().hex, 'name': uuid.uuid4().hex}
        PROVIDERS.resource_api.create_domain(domain2['id'], domain2)
        PROVIDERS.domain_config_api.release_registration(
            domain2['id'], type=type
        )

        # If nobody has the type registered, then trying to read it should
        # raise ConfigRegistrationNotFound
        PROVIDERS.domain_config_api.release_registration(
            self.domain['id'], type=type
        )
        self.assertRaises(
            exception.ConfigRegistrationNotFound,
            PROVIDERS.domain_config_api.read_registration,
            type,
        )

        # Finally check multiple registrations are cleared if you free the
        # registration without specifying the type
        type2 = uuid.uuid4().hex
        PROVIDERS.domain_config_api.obtain_registration(
            self.domain['id'], type
        )
        PROVIDERS.domain_config_api.obtain_registration(
            self.domain['id'], type2
        )
        PROVIDERS.domain_config_api.release_registration(self.domain['id'])
        self.assertRaises(
            exception.ConfigRegistrationNotFound,
            PROVIDERS.domain_config_api.read_registration,
            type,
        )
        self.assertRaises(
            exception.ConfigRegistrationNotFound,
            PROVIDERS.domain_config_api.read_registration,
            type2,
        )

    def test_option_dict_fails_when_group_is_none(self):
        group = 'foo'
        option = 'bar'
        self.assertRaises(
            cfg.NoSuchOptError,
            PROVIDERS.domain_config_api._option_dict,
            group,
            option,
        )

    def test_option_dict_returns_valid_config_values(self):
        regex = uuid.uuid4().hex
        self.config_fixture.config(
            group='security_compliance', password_regex=regex
        )
        expected_dict = {
            'group': 'security_compliance',
            'option': 'password_regex',
            'value': regex,
        }
        option_dict = PROVIDERS.domain_config_api._option_dict(
            'security_compliance', 'password_regex'
        )
        self.assertEqual(option_dict, expected_dict)



# Copyright 2018 Catalyst Cloud Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import datetime
import struct
import uuid

from cryptography import fernet
import msgpack
from oslo_log import log
from oslo_utils import timeutils

from keystone.auth import plugins as auth_plugins
from keystone.common import fernet_utils as utils
from keystone.common import utils as ks_utils
import keystone.conf
from keystone import exception
from keystone.i18n import _

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)

# Fernet byte indexes as computed by pypi/keyless_fernet and defined in
# https://github.com/fernet/spec
TIMESTAMP_START = 1
TIMESTAMP_END = 9


class ReceiptFormatter:
    """Packs and unpacks payloads into receipts for transport."""

    @property
    def crypto(self):
        """Return a cryptography instance.

        You can extend this class with a custom crypto @property to provide
        your own receipt encoding / decoding. For example, using a different
        cryptography library (e.g. ``python-keyczar``) or to meet arbitrary
        security requirements.

        This @property just needs to return an object that implements
        ``encrypt(plaintext)`` and ``decrypt(ciphertext)``.

        """
        fernet_utils = utils.FernetUtils(
            CONF.fernet_receipts.key_repository,
            CONF.fernet_receipts.max_active_keys,
            'fernet_receipts',
        )
        keys = fernet_utils.load_keys()

        if not keys:
            raise exception.KeysNotFound()

        fernet_instances = [fernet.Fernet(key) for key in keys]
        return fernet.MultiFernet(fernet_instances)

    def pack(self, payload):
        """Pack a payload for transport as a receipt.

        :type payload: bytes
        :rtype: str

        """
        # base64 padding (if any) is not URL-safe
        return self.crypto.encrypt(payload).rstrip(b'=').decode('utf-8')

    def unpack(self, receipt):
        """Unpack a receipt, and validate the payload.

        :type receipt: str
        :rtype: bytes

        """
        receipt = ReceiptFormatter.restore_padding(receipt)

        try:
            return self.crypto.decrypt(receipt.encode('utf-8'))
        except fernet.InvalidToken:
            raise exception.ValidationError(
                _('This is not a recognized Fernet receipt %s') % receipt
            )

    @classmethod
    def restore_padding(cls, receipt):
        """Restore padding based on receipt size.

        :param receipt: receipt to restore padding on
        :type receipt: str
        :returns: receipt with correct padding

        """
        # Re-inflate the padding
        mod_returned = len(receipt) % 4
        if mod_returned:
            missing_padding = 4 - mod_returned
            receipt += '=' * missing_padding
        return receipt

    @classmethod
    def creation_time(cls, fernet_receipt):
        """Return the creation time of a valid Fernet receipt.

        :type fernet_receipt: str

        """
        fernet_receipt = ReceiptFormatter.restore_padding(fernet_receipt)
        # fernet_receipt is str

        # Fernet receipts are base64 encoded, so we need to unpack them first
        # urlsafe_b64decode() requires bytes
        receipt_bytes = base64.urlsafe_b64decode(
            fernet_receipt.encode('utf-8')
        )

        # slice into the byte array to get just the timestamp
        timestamp_bytes = receipt_bytes[TIMESTAMP_START:TIMESTAMP_END]

        # convert those bytes to an integer
        # (it's a 64-bit "unsigned long long int" in C)
        timestamp_int = struct.unpack(">Q", timestamp_bytes)[0]

        # and with an integer, it's trivial to produce a datetime object
        issued_at = datetime.datetime.fromtimestamp(
            timestamp_int, datetime.timezone.utc
        ).replace(tzinfo=None)

        return issued_at

    def create_receipt(self, user_id, methods, expires_at):
        """Given a set of payload attributes, generate a Fernet receipt."""
        payload = ReceiptPayload.assemble(user_id, methods, expires_at)

        serialized_payload = msgpack.packb(payload)
        receipt = self.pack(serialized_payload)

        # NOTE(lbragstad): We should warn against Fernet receipts that are over
        # 255 characters in length. This is mostly due to persisting the
        # receipts in a backend store of some kind that might have a limit of
        # 255 characters. Even though Keystone isn't storing a Fernet receipt
        # anywhere, we can't say it isn't being stored somewhere else with
        # those kind of backend constraints.
        if len(receipt) > 255:
            LOG.info(
                'Fernet receipt created with length of %d '
                'characters, which exceeds 255 characters',
                len(receipt),
            )

        return receipt

    def validate_receipt(self, receipt):
        """Validate a Fernet receipt and returns the payload attributes.

        :type receipt: str

        """
        serialized_payload = self.unpack(receipt)
        payload = msgpack.unpackb(serialized_payload)

        (user_id, methods, expires_at) = ReceiptPayload.disassemble(payload)

        # rather than appearing in the payload, the creation time is encoded
        # into the receipt format itself
        issued_at = ReceiptFormatter.creation_time(receipt)
        issued_at = ks_utils.isotime(at=issued_at, subsecond=True)
        expires_at = timeutils.parse_isotime(expires_at)
        expires_at = ks_utils.isotime(at=expires_at, subsecond=True)

        return (user_id, methods, issued_at, expires_at)


class ReceiptPayload:
    @classmethod
    def assemble(cls, user_id, methods, expires_at):
        """Assemble the payload of a receipt.

        :param user_id: identifier of the user in the receipt request
        :param methods: list of authentication methods used
        :param expires_at: datetime of the receipt's expiration
        :returns: the payload of a receipt

        """
        b_user_id = cls.attempt_convert_uuid_hex_to_bytes(user_id)
        methods = auth_plugins.convert_method_list_to_integer(methods)
        expires_at_int = cls._convert_time_string_to_float(expires_at)
        return (b_user_id, methods, expires_at_int)

    @classmethod
    def disassemble(cls, payload):
        """Disassemble a payload into the component data.

        The tuple consists of::

            (user_id, methods, expires_at_str)

        * ``methods`` are the auth methods.

        :param payload: this variant of payload
        :returns: a tuple of the payloads component data

        """
        (is_stored_as_bytes, user_id) = payload[0]
        if is_stored_as_bytes:
            user_id = cls.convert_uuid_bytes_to_hex(user_id)
        methods = auth_plugins.convert_integer_to_method_list(payload[1])
        expires_at_str = cls._convert_float_to_time_string(payload[2])
        return (user_id, methods, expires_at_str)

    @classmethod
    def convert_uuid_hex_to_bytes(cls, uuid_string):
        """Compress UUID formatted strings to bytes.

        :param uuid_string: uuid string to compress to bytes
        :returns: a byte representation of the uuid

        """
        uuid_obj = uuid.UUID(uuid_string)
        return uuid_obj.bytes

    @classmethod
    def convert_uuid_bytes_to_hex(cls, uuid_byte_string):
        """Generate uuid.hex format based on byte string.

        :param uuid_byte_string: uuid string to generate from
        :returns: uuid hex formatted string

        """
        uuid_obj = uuid.UUID(bytes=uuid_byte_string)
        return uuid_obj.hex

    @classmethod
    def _convert_time_string_to_float(cls, time_string):
        """Convert a time formatted string to a float.

        :param time_string: time formatted string
        :returns: a timestamp as a float

        """
        time_object = timeutils.parse_isotime(time_string)
        return (
            timeutils.normalize_time(time_object)
            - datetime.datetime.fromtimestamp(
                0, datetime.timezone.utc
            ).replace(tzinfo=None)
        ).total_seconds()

    @classmethod
    def _convert_float_to_time_string(cls, time_float):
        """Convert a floating point timestamp to a string.

        :param time_float: integer representing timestamp
        :returns: a time formatted strings

        """
        time_object = datetime.datetime.fromtimestamp(
            time_float, datetime.timezone.utc
        ).replace(tzinfo=None)
        return ks_utils.isotime(time_object, subsecond=True)

    @classmethod
    def attempt_convert_uuid_hex_to_bytes(cls, value):
        """Attempt to convert value to bytes or return value.

        :param value: value to attempt to convert to bytes
        :returns: tuple containing boolean indicating whether user_id was
                  stored as bytes and uuid value as bytes or the original value

        """
        try:
            return (True, cls.convert_uuid_hex_to_bytes(value))
        except ValueError:
            # this might not be a UUID, depending on the situation (i.e.
            # federation)
            return (False, value)

    @classmethod
    def base64_encode(cls, s):
        """Encode a URL-safe string.

        :type s: str
        :rtype: str

        """
        # urlsafe_b64encode() returns bytes so need to convert to
        # str, might as well do it before stripping.
        return base64.urlsafe_b64encode(s).decode('utf-8').rstrip('=')

    @classmethod
    def random_urlsafe_str_to_bytes(cls, s):
        """Convert string from :func:`random_urlsafe_str()` to bytes.

        :type s: str
        :rtype: bytes

        """
        # urlsafe_b64decode() requires str, unicode isn't accepted.
        s = str(s)

        # restore the padding (==) at the end of the string
        return base64.urlsafe_b64decode(s + '==')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This file handles all flask-restful resources for /v3/registered_limits

import http.client

import flask

from keystone.api import validation
from keystone.common import json_home
from keystone.common import provider_api
from keystone.common import rbac_enforcer
from keystone.limit import schema
from keystone.server import flask as ks_flask

PROVIDERS = provider_api.ProviderAPIs
ENFORCER = rbac_enforcer.RBACEnforcer


class RegisteredLimitsResource(ks_flask.ResourceBase):
    collection_key = 'registered_limits'
    member_key = 'registered_limit'
    json_home_resource_status = json_home.Status.EXPERIMENTAL

    @validation.request_query_schema(
        schema.registered_limits_index_request_query
    )
    @validation.response_body_schema(
        schema.registered_limits_index_response_body
    )
    def get(self):
        """List registered limits.

        GET /v3/registered_limits
        """
        filters = ['service_id', 'region_id', 'resource_name']
        ENFORCER.enforce_call(
            action='identity:list_registered_limits', filters=filters
        )
        hints = self.build_driver_hints(filters)
        refs = PROVIDERS.unified_limit_api.list_registered_limits(hints)
        return self.wrap_collection(refs, hints=hints)

    @validation.request_body_schema(
        schema.registered_limits_create_request_body
    )
    @validation.response_body_schema(
        schema.registered_limits_create_response_body
    )
    def post(self):
        """Create new registered limits.

        POST /v3/registered_limits
        """
        ENFORCER.enforce_call(action='identity:create_registered_limits')
        reg_limits = (
            flask.request.get_json(silent=True, force=True) or {}
        ).get('registered_limits', {})
        registered_limits = [
            self._assign_unique_id(self._normalize_dict(r)) for r in reg_limits
        ]
        refs = PROVIDERS.unified_limit_api.create_registered_limits(
            registered_limits
        )
        refs = self.wrap_collection(refs)
        refs.pop('links')
        return refs, http.client.CREATED


class RegisteredLimitResource(ks_flask.ResourceBase):
    collection_key = 'registered_limits'
    member_key = 'registered_limit'
    json_home_resource_status = json_home.Status.EXPERIMENTAL

    @validation.request_body_schema(None)
    @validation.response_body_schema(
        schema.registered_limit_show_response_body
    )
    def get(self, registered_limit_id=None):
        """Retrieve an existing registered limit.

        GET /v3/registered_limits/{registered_limit_id}
        """
        ENFORCER.enforce_call(action='identity:get_registered_limit')
        ref = PROVIDERS.unified_limit_api.get_registered_limit(
            registered_limit_id
        )
        return self.wrap_member(ref)

    @validation.request_body_schema(
        schema.registered_limit_update_request_body
    )
    @validation.response_body_schema(
        schema.registered_limit_show_response_body
    )
    def patch(self, registered_limit_id):
        """Update an existing registered limit.

        PATCH /v3/registered_limits/{registered_limit_id}
        """
        ENFORCER.enforce_call(action='identity:update_registered_limit')
        registered_limit = (
            flask.request.get_json(silent=True, force=True) or {}
        ).get('registered_limit', {})
        self._require_matching_id(registered_limit)
        ref = PROVIDERS.unified_limit_api.update_registered_limit(
            registered_limit_id, registered_limit
        )
        return self.wrap_member(ref)

    @validation.request_body_schema(None)
    @validation.response_body_schema(None)
    def delete(self, registered_limit_id):
        """Delete a registered limit.

        DELETE /v3/registered_limits/{registered_limit_id}
        """
        ENFORCER.enforce_call(action='identity:delete_registered_limit')
        return (
            PROVIDERS.unified_limit_api.delete_registered_limit(
                registered_limit_id
            ),
            http.client.NO_CONTENT,
        )


class RegisteredLimitsAPI(ks_flask.APIBase):
    _name = 'registered_limits'
    _import_name = __name__
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=RegisteredLimitsResource,
            url='/registered_limits',
            resource_kwargs={},
            rel="registered_limits",
            path_vars=None,
            status=json_home.Status.EXPERIMENTAL,
        ),
        ks_flask.construct_resource_map(
            resource=RegisteredLimitResource,
            url='/registered_limits/<string:registered_limit_id>',
            resource_kwargs={},
            rel="registered_limit",
            path_vars={
                'registered_limit_id': json_home.build_v3_parameter_relation(
                    "registered_limit_id"
                )
            },
            status=json_home.Status.EXPERIMENTAL,
        ),
    ]


APIs = (RegisteredLimitsAPI,)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Common parameter types for validating API requests."""

from typing import Any

empty: dict[str, Any] = {"type": "null"}

name: dict[str, Any] = {
    "type": "string",
    "minLength": 1,
    "maxLength": 255,
    "pattern": r"[\S]+",
    "description": "The resource name.",
}

boolean = {
    "type": ["boolean", "string", "null"],
    "enum": [
        True,
        "True",
        "TRUE",
        "true",
        False,
        "False",
        "FALSE",
        "false",
        "",
        "1",
        "0",
        "y",
        "Y",
        "n",
        "N",
        "on",
        "ON",
        "off",
        "OFF",
        "yes",
        "no",
    ],
}

description: dict[str, Any] = {
    "type": "string",
    "description": "The resource description.",
}

domain_id: dict[str, Any] = {
    "type": "string",
    "minLength": 1,
    "maxLength": 64,
    "pattern": r"^[a-zA-Z0-9-]+$",
    "description": "The ID of the domain.",
}

project_id: dict[str, Any] = {
    "type": "string",
    "minLength": 1,
    "maxLength": 64,
    "pattern": r"^[a-zA-Z0-9-]+$",
    "description": "The ID of the project.",
}

parent_id: dict[str, str] = {"type": "string", "format": "uuid"}

region_id: dict[str, Any] = {
    "type": ["string", "null"],
    "minLength": 1,
    "maxLength": 255,
    "description": "The ID of the region.",
}

_tag_name_property = {
    "type": "string",
    "minLength": 1,
    "maxLength": 255,
    # NOTE(gagehugo) This pattern is for tags which follows the
    # guidelines as set by the API-WG, which matches anything that
    # does not contain a '/' or ','.
    # https://specs.openstack.org/openstack/api-wg/guidelines/tags.html
    "pattern": r"^[^,/]*$",
}

tags: dict[str, Any] = {
    "type": "string",
    "x-openstack": {
        # As OpenAPI request parameters this is an array of string serialized
        # as csv
        "openapi": {
            "schema": {"type": "array", "items": _tag_name_property},
            "style": "form",
            "explode": False,
        }
    },
}

url: dict[str, Any] = {
    "type": "string",
    "minLength": 0,
    "maxLength": 225,
    "pattern": "^[a-zA-Z0-9+.-]+:.+",
}

id_string: dict[str, Any] = {
    "type": "string",
    "minLength": 1,
    "maxLength": 64,
    "pattern": r"^[a-zA-Z0-9-]+$",
}

sort_key: dict[str, Any] = {
    "type": "string",
    "description": "Sorts resources by attribute.",
}

sort_dir: dict[str, Any] = {
    "type": "string",
    "description": "Sort direction. A valid value is asc (ascending) or desc (descending).",
    "enum": ["asc", "desc"],
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_db import api as oslo_db_api
import sqlalchemy

from keystone.common import sql
from keystone.models import revoke_model
from keystone.revoke.backends import base


class RevocationEvent(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'revocation_event'
    attributes = revoke_model.REVOKE_KEYS

    # The id field is not going to be exposed to the outside world.
    # It is, however, necessary for SQLAlchemy.
    id = sql.Column(sql.Integer, primary_key=True, nullable=False)
    domain_id = sql.Column(sql.String(64))
    project_id = sql.Column(sql.String(64))
    user_id = sql.Column(sql.String(64))
    role_id = sql.Column(sql.String(64))
    trust_id = sql.Column(sql.String(64))
    consumer_id = sql.Column(sql.String(64))
    access_token_id = sql.Column(sql.String(64))
    issued_before = sql.Column(sql.DateTime(), nullable=False, index=True)
    expires_at = sql.Column(sql.DateTime())
    revoked_at = sql.Column(sql.DateTime(), nullable=False, index=True)
    audit_id = sql.Column(sql.String(32))
    audit_chain_id = sql.Column(sql.String(32))
    __table_args__ = (
        sql.Index(
            'ix_revocation_event_project_id_issued_before',
            'project_id',
            'issued_before',
        ),
        sql.Index(
            'ix_revocation_event_user_id_issued_before',
            'user_id',
            'issued_before',
        ),
        sql.Index(
            'ix_revocation_event_audit_id_issued_before',
            'audit_id',
            'issued_before',
        ),
    )


class Revoke(base.RevokeDriverBase):
    def _flush_batch_size(self, dialect):
        batch_size = 0
        if dialect == 'ibm_db_sa':
            # This functionality is limited to DB2, because
            # it is necessary to prevent the transaction log
            # from filling up, whereas at least some of the
            # other supported databases do not support update
            # queries with LIMIT subqueries nor do they appear
            # to require the use of such queries when deleting
            # large numbers of records at once.
            batch_size = 100
            # Limit of 100 is known to not fill a transaction log
            # of default maximum size while not significantly
            # impacting the performance of large token purges on
            # systems where the maximum transaction log size has
            # been increased beyond the default.
        return batch_size

    def _prune_expired_events(self):
        oldest = base.revoked_before_cutoff_time()

        with sql.session_for_write() as session:
            dialect = session.bind.dialect.name
            batch_size = self._flush_batch_size(dialect)
            if batch_size > 0:
                query = session.query(RevocationEvent.id)
                query = query.filter(RevocationEvent.revoked_at < oldest)
                query = query.limit(batch_size).subquery()
                delete_query = session.query(RevocationEvent).filter(
                    RevocationEvent.id.in_(query)
                )
                while True:
                    rowcount = delete_query.delete(synchronize_session=False)
                    if rowcount == 0:
                        break
            else:
                query = session.query(RevocationEvent)
                query = query.filter(RevocationEvent.revoked_at < oldest)
                query.delete(synchronize_session=False)

            session.flush()

    def _list_token_events(self, token):
        with sql.session_for_read() as session:
            query = session.query(RevocationEvent).filter(
                RevocationEvent.issued_before >= token['issued_at']
            )
            user = [RevocationEvent.user_id.is_(None)]
            proj = [RevocationEvent.project_id.is_(None)]
            audit = [RevocationEvent.audit_id.is_(None)]
            trust = [RevocationEvent.trust_id.is_(None)]
            if token['user_id']:
                user.append(RevocationEvent.user_id == token['user_id'])
            if token['trustor_id']:
                user.append(RevocationEvent.user_id == token['trustor_id'])
            if token['trustee_id']:
                user.append(RevocationEvent.user_id == token['trustee_id'])
            if token['project_id']:
                proj.append(RevocationEvent.project_id == token['project_id'])
            if token['audit_id']:
                audit.append(RevocationEvent.audit_id == token['audit_id'])
            if token['trust_id']:
                trust.append(RevocationEvent.trust_id == token['trust_id'])
            query = query.filter(
                sqlalchemy.and_(
                    sqlalchemy.or_(*user),
                    sqlalchemy.or_(*proj),
                    sqlalchemy.or_(*audit),
                    sqlalchemy.or_(*trust),
                )
            )
            events = [revoke_model.RevokeEvent(**e.to_dict()) for e in query]
            return events

    def _list_last_fetch_events(self, last_fetch=None):
        with sql.session_for_read() as session:
            query = session.query(RevocationEvent).order_by(
                RevocationEvent.revoked_at
            )

            if last_fetch:
                query = query.filter(RevocationEvent.revoked_at > last_fetch)

            events = [revoke_model.RevokeEvent(**e.to_dict()) for e in query]
            return events

    def list_events(self, last_fetch=None, token=None):
        if token:
            return self._list_token_events(token)
        else:
            return self._list_last_fetch_events(last_fetch)

    @oslo_db_api.wrap_db_retry(retry_on_deadlock=True)
    def revoke(self, event):
        kwargs = {}
        for attr in revoke_model.REVOKE_KEYS:
            kwargs[attr] = getattr(event, attr)
        record = RevocationEvent(**kwargs)
        with sql.session_for_write() as session:
            session.add(record)
            self._prune_expired_events()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Remove service_provider.relay_state_prefix server default.

Revision ID: 11c3b243b4cb
Revises: b4f8b3f584e0
Create Date: 2023-07-03 12:03:21.649144
"""

from alembic import op

# revision identifiers, used by Alembic.
revision = '11c3b243b4cb'
down_revision = 'b4f8b3f584e0'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('service_provider', schema=None) as batch_op:
        batch_op.alter_column('relay_state_prefix', server_default=None)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Unit tests for core identity behavior."""

import itertools
import os
from unittest import mock
import uuid

import fixtures
from oslo_config import fixture as config_fixture
import stevedore

from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone import identity
from keystone.tests import unit
from keystone.tests.unit import default_fixtures
from keystone.tests.unit.identity.backends import fake_driver
from keystone.tests.unit.ksfixtures import database

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class TestDomainConfigs(unit.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.addCleanup(CONF.reset)

        self.tmp_dir = unit.dirs.tmp()

        self.config_fixture = self.useFixture(config_fixture.Config(CONF))
        self.config_fixture.config(
            domain_config_dir=self.tmp_dir, group='identity'
        )

    def test_config_for_nonexistent_domain(self):
        """Having a config for a non-existent domain will be ignored.

        There are no assertions in this test because there are no side
        effects. If there is a config file for a domain that does not
        exist it should be ignored.

        """
        domain_id = uuid.uuid4().hex
        domain_config_filename = os.path.join(
            self.tmp_dir, f'keystone.{domain_id}.conf'
        )
        self.addCleanup(lambda: os.remove(domain_config_filename))
        with open(domain_config_filename, 'w'):
            """Write an empty config file."""

        e = exception.DomainNotFound(domain_id=domain_id)
        mock_assignment_api = mock.Mock()
        mock_assignment_api.get_domain_by_name.side_effect = e

        domain_config = identity.DomainConfigs()
        fake_standard_driver = None
        domain_config.setup_domain_drivers(
            fake_standard_driver, mock_assignment_api
        )

    def test_config_for_dot_name_domain(self):
        # Ensure we can get the right domain name which has dots within it
        # from filename.
        domain_config_filename = os.path.join(
            self.tmp_dir, 'keystone.abc.def.com.conf'
        )
        with open(domain_config_filename, 'w'):
            """Write an empty config file."""
        self.addCleanup(os.remove, domain_config_filename)

        with mock.patch.object(
            identity.DomainConfigs, '_load_config_from_file'
        ) as mock_load_config:
            domain_config = identity.DomainConfigs()
            fake_assignment_api = None
            fake_standard_driver = None
            domain_config.setup_domain_drivers(
                fake_standard_driver, fake_assignment_api
            )
            mock_load_config.assert_called_once_with(
                fake_assignment_api, [domain_config_filename], 'abc.def.com'
            )

    def test_config_for_multiple_sql_backend(self):
        domains_config = identity.DomainConfigs()

        # Create the right sequence of is_sql in the drivers being
        # requested to expose the bug, which is that a False setting
        # means it forgets previous True settings.
        drivers = []
        files = []
        for idx, is_sql in enumerate((True, False, True)):
            drv = mock.Mock(is_sql=is_sql)
            drivers.append(drv)
            name = f'dummy.{idx}'
            files.append(
                ''.join(
                    (
                        identity.DOMAIN_CONF_FHEAD,
                        name,
                        identity.DOMAIN_CONF_FTAIL,
                    )
                )
            )

        def walk_fake(*a, **kwa):
            return (('/fake/keystone/domains/config', [], files),)

        generic_driver = mock.Mock(is_sql=False)

        assignment_api = mock.Mock()
        id_factory = itertools.count()
        assignment_api.get_domain_by_name.side_effect = lambda name: {
            'id': next(id_factory),
            '_': 'fake_domain',
        }
        load_driver_mock = mock.Mock(side_effect=drivers)

        with mock.patch.object(os, 'walk', walk_fake):
            with mock.patch.object(identity.cfg, 'ConfigOpts'):
                with mock.patch.object(
                    domains_config, '_load_driver', load_driver_mock
                ):
                    self.assertRaises(
                        exception.MultipleSQLDriversInConfig,
                        domains_config.setup_domain_drivers,
                        generic_driver,
                        assignment_api,
                    )

                    self.assertEqual(3, load_driver_mock.call_count)


class TestDatabaseDomainConfigs(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(database.Database())
        self.load_backends()
        PROVIDERS.resource_api.create_domain(
            default_fixtures.ROOT_DOMAIN['id'], default_fixtures.ROOT_DOMAIN
        )

    def test_domain_config_in_database_disabled_by_default(self):
        self.assertFalse(CONF.identity.domain_configurations_from_database)

    def test_loading_config_from_database(self):
        self.config_fixture.config(
            domain_configurations_from_database=True, group='identity'
        )
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        # Override two config options for our domain
        conf = {
            'ldap': {
                'url': uuid.uuid4().hex,
                'suffix': uuid.uuid4().hex,
                'use_tls': True,
            },
            'identity': {'driver': 'ldap'},
        }
        PROVIDERS.domain_config_api.create_config(domain['id'], conf)
        fake_standard_driver = None
        domain_config = identity.DomainConfigs()
        domain_config.setup_domain_drivers(
            fake_standard_driver, PROVIDERS.resource_api
        )
        # Make sure our two overrides are in place, and others are not affected
        res = domain_config.get_domain_conf(domain['id'])
        self.assertEqual(conf['ldap']['url'], res.ldap.url)
        self.assertEqual(conf['ldap']['suffix'], res.ldap.suffix)
        self.assertEqual(CONF.ldap.query_scope, res.ldap.query_scope)

        # Make sure the override is not changing the type of the config value
        use_tls_type = type(CONF.ldap.use_tls)
        self.assertEqual(
            use_tls_type(conf['ldap']['use_tls']), res.ldap.use_tls
        )

        # Now turn off using database domain configuration and check that the
        # default config file values are now seen instead of the overrides.
        self.config_fixture.config(
            group='identity', domain_configurations_from_database=False
        )

        domain_config = identity.DomainConfigs()
        domain_config.setup_domain_drivers(
            fake_standard_driver, PROVIDERS.resource_api
        )
        res = domain_config.get_domain_conf(domain['id'])
        self.assertEqual(CONF.ldap.url, res.ldap.url)
        self.assertEqual(CONF.ldap.suffix, res.ldap.suffix)
        self.assertEqual(CONF.ldap.use_tls, res.ldap.use_tls)
        self.assertEqual(CONF.ldap.query_scope, res.ldap.query_scope)

    def test_loading_config_from_database_out_of_tree(self):
        # Test domain config loading for out-of-tree driver supporting own
        # config options

        # Prepare fake driver
        extension = stevedore.extension.Extension(
            name="foo",
            entry_point=None,
            obj=fake_driver.FooDriver(),
            plugin=None,
        )
        fake_driver_manager = stevedore.DriverManager.make_test_instance(
            extension, namespace="keystone.identity"
        )
        # replace DriverManager with a patched test instance
        self.useFixture(
            fixtures.MockPatchObject(
                stevedore, "DriverManager", return_value=fake_driver_manager
            )
        ).mock

        self.config_fixture.config(
            domain_configurations_from_database=True, group="identity"
        )
        self.config_fixture.config(
            additional_whitelisted_options={"foo": ["opt1"]},
            group="domain_config",
        )
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain["id"], domain)
        # Override two config options for our domain
        conf = {
            "foo": {"opt1": uuid.uuid4().hex},
            "identity": {"driver": "foo"},
        }
        PROVIDERS.domain_config_api.create_config(domain["id"], conf)
        domain_config = identity.DomainConfigs()
        domain_config.setup_domain_drivers("foo", PROVIDERS.resource_api)
        # Make sure our two overrides are in place, and others are not affected
        res = domain_config.get_domain_conf(domain["id"])

        self.assertEqual(conf["foo"]["opt1"], res.foo.opt1)

        # Reset whitelisted options in the provider directly. Due to the fact
        # that there are too many singletons used around the code basis there
        # is a chance of clash when other API domain_config tests are being
        # executed by the same process. It is NOT ENOUGH just to invoke reset
        # on fixtures.
        PROVIDERS.domain_config_api.whitelisted_options.pop("foo", None)



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import inspect
import time
import types

from oslo_log import log
import stevedore

from keystone.common.driver_hints import Hints
from keystone.common import provider_api
from keystone.i18n import _

LOG = log.getLogger(__name__)


def response_truncated(f):
    """Truncate the list returned by the wrapped function.

    This is designed to wrap Manager list_{entity} methods to ensure that
    any list limits that are defined are passed to the driver layer.  If a
    hints list is provided, the wrapper will insert the relevant limit into
    the hints so that the underlying driver call can try and honor it. If the
    driver does truncate the response, it will update the 'truncated' attribute
    in the 'limit' entry in the hints list, which enables the caller of this
    function to know if truncation has taken place.  If, however, the driver
    layer is unable to perform truncation, the 'limit' entry is simply left in
    the hints list for the caller to handle.

    A _get_list_limit() method is required to be present in the object class
    hierarchy, which returns the limit for this backend to which we will
    truncate.

    If a hints list is not provided in the arguments of the wrapped call then
    any limits set in the config file are ignored.  This allows internal use
    of such wrapped methods where the entire data set is needed as input for
    the calculations of some other API (e.g. get role assignments for a given
    project).

    """

    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        if kwargs.get('hints') is None:
            return f(self, *args, **kwargs)

        # Set default limit if not there already
        if (
            isinstance(kwargs['hints'], Hints)
            and kwargs['hints'].limit is None
        ):
            list_limit = self.driver._get_list_limit()
            if list_limit:
                kwargs['hints'].set_limit(list_limit)
        return f(self, *args, **kwargs)

    return wrapper


def load_driver(namespace, driver_name, *args):
    try:
        driver_manager = stevedore.DriverManager(
            namespace, driver_name, invoke_on_load=True, invoke_args=args
        )
        return driver_manager.driver
    except stevedore.exception.NoMatches:
        msg = _('Unable to find %(name)r driver in %(namespace)r.')
        raise ImportError(msg % {'name': driver_name, 'namespace': namespace})


class _TraceMeta(type):
    """A metaclass that, in trace mode, will log entry and exit of methods.

    This metaclass automatically wraps all methods on the class when
    instantiated with a decorator that will log entry/exit from a method
    when keystone is run in Trace log level.
    """

    @staticmethod
    def wrapper(__f, __classname):
        __argspec = inspect.getfullargspec(__f)
        __fn_info = (
            f'{inspect.getmodule(__f).__name__}.{__classname}.{__f.__name__}'
        )
        # NOTE(morganfainberg): Omit "cls" and "self" when printing trace logs
        # the index can be calculated at wrap time rather than at runtime.
        if __argspec.args and __argspec.args[0] in ('self', 'cls'):
            __arg_idx = 1
        else:
            __arg_idx = 0

        @functools.wraps(__f)
        def wrapped(*args, **kwargs):
            __exc = None
            __t = time.time()
            __do_trace = LOG.logger.getEffectiveLevel() <= log.TRACE
            __ret_val = None
            try:
                if __do_trace:
                    LOG.trace('CALL => %s', __fn_info)
                __ret_val = __f(*args, **kwargs)
            except Exception as e:  # nosec
                __exc = e
                raise
            finally:
                if __do_trace:
                    __subst = {
                        'run_time': (time.time() - __t),
                        'passed_args': ', '.join(
                            [
                                ', '.join([repr(a) for a in args[__arg_idx:]]),
                                ', '.join(
                                    [f'{k}={v!r}' for k, v in kwargs.items()]
                                ),
                            ]
                        ),
                        'function': __fn_info,
                        'exception': __exc,
                        'ret_val': __ret_val,
                    }
                    if __exc is not None:
                        __msg = (
                            '[%(run_time)ss] %(function)s '
                            '(%(passed_args)s) => raised '
                            '%(exception)r'
                        )
                    else:
                        # TODO(morganfainberg): find a way to indicate if this
                        # was a cache hit or cache miss.
                        __msg = (
                            '[%(run_time)ss] %(function)s'
                            '(%(passed_args)s) => %(ret_val)r'
                        )
                    LOG.trace(__msg, __subst)
            return __ret_val

        return wrapped

    def __new__(meta, classname, bases, class_dict):
        final_cls_dict = {}
        for attr_name, attr in class_dict.items():
            # NOTE(morganfainberg): only wrap public instances and methods.
            if isinstance(
                attr, types.FunctionType
            ) and not attr_name.startswith('_'):
                attr = _TraceMeta.wrapper(attr, classname)
            final_cls_dict[attr_name] = attr
        return type.__new__(meta, classname, bases, final_cls_dict)


class Manager(metaclass=_TraceMeta):
    """Base class for intermediary request layer.

    The Manager layer exists to support additional logic that applies to all
    or some of the methods exposed by a service that are not specific to the
    HTTP interface.

    It also provides a stable entry point to dynamic backends.

    An example of a probable use case is logging all the calls.

    """

    driver_namespace: str
    _provides_api: str

    def __init__(self, driver_name):
        if self._provides_api is None:
            raise ValueError(
                'Programming Error: All managers must provide an '
                'API that can be referenced by other components '
                'of Keystone.'
            )
        if driver_name is not None:
            self.driver = load_driver(self.driver_namespace, driver_name)
        self.__register_provider_api()

    def __register_provider_api(self):
        provider_api.ProviderAPIs._register_provider_api(
            name=self._provides_api, obj=self
        )

    def __getattr__(self, name):
        """Forward calls to the underlying driver.

        This method checks for a provider api before forwarding.
        """
        try:
            return getattr(provider_api.ProviderAPIs, name)
        except AttributeError:
            # NOTE(morgan): We didn't find a provider api, move on and
            # forward to the driver as expected.
            pass

        f = getattr(self.driver, name)
        if callable(f):
            # NOTE(dstanek): only if this is callable (class or function)
            # cache this
            setattr(self, name, f)
        return f



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import fixtures
import freezegun
from oslo_config import fixture as config_fixture
from oslo_log import log
from oslo_utils import timeutils

from keystone.common import fernet_utils
from keystone.common import utils as common_utils
import keystone.conf
from keystone.credential.providers import fernet as credential_fernet
from keystone import exception
from keystone.server.flask import application
from keystone.tests import unit
from keystone.tests.unit import ksfixtures
from keystone.tests.unit import utils

CONF = keystone.conf.CONF

TZ = utils.TZ


class UtilsTestCase(unit.BaseTestCase):
    OPTIONAL = object()

    def setUp(self):
        super().setUp()
        self.config_fixture = self.useFixture(config_fixture.Config(CONF))

    def test_resource_uuid(self):
        # Basic uuid test, most IDs issued by keystone look like this:
        value = '536e28c2017e405e89b25a1ed777b952'
        self.assertEqual(value, common_utils.resource_uuid(value))

    def test_resource_64_char_uuid(self):
        # Exact 64 length string, like ones used by mapping_id backend, are not
        # valid UUIDs, so they will be UUID5 namespaced
        value = 'f13de678ac714bb1b7d1e9a007c10db5' * 2
        expected_id = uuid.uuid5(common_utils.RESOURCE_ID_NAMESPACE, value).hex
        self.assertEqual(expected_id, common_utils.resource_uuid(value))

    def test_resource_non_ascii_chars(self):
        # IDs with non-ASCII characters will be UUID5 namespaced
        value = 'ß' * 32
        expected_id = uuid.uuid5(common_utils.RESOURCE_ID_NAMESPACE, value).hex
        self.assertEqual(expected_id, common_utils.resource_uuid(value))

    def test_resource_invalid_id(self):
        # This input is invalid because it's length is more than 64.
        value = 'x' * 65
        self.assertRaises(ValueError, common_utils.resource_uuid, value)

    def test_hash(self):
        password = 'right'
        wrong = 'wrongwrong'  # Two wrongs don't make a right
        hashed = common_utils.hash_password(password)
        self.assertTrue(common_utils.check_password(password, hashed))
        self.assertFalse(common_utils.check_password(wrong, hashed))

    def test_verify_normal_password_strict(self):
        self.config_fixture.config(strict_password_check=False)
        password = uuid.uuid4().hex
        verified = common_utils.verify_length_and_trunc_password(password)
        self.assertEqual(password.encode('utf-8'), verified)

    def test_that_a_hash_can_not_be_validated_against_a_hash(self):
        # NOTE(dstanek): Bug 1279849 reported a problem where passwords
        # were not being hashed if they already looked like a hash. This
        # would allow someone to hash their password ahead of time
        # (potentially getting around password requirements, like
        # length) and then they could auth with their original password.
        password = uuid.uuid4().hex
        hashed_password = common_utils.hash_password(password)
        new_hashed_password = common_utils.hash_password(hashed_password)
        self.assertFalse(
            common_utils.check_password(password, new_hashed_password)
        )

    def test_verify_long_password_strict(self):
        self.config_fixture.config(strict_password_check=False)
        self.config_fixture.config(group='identity', max_password_length=5)
        max_length = CONF.identity.max_password_length
        invalid_password = 'passw0rd'
        trunc = common_utils.verify_length_and_trunc_password(invalid_password)
        self.assertEqual(invalid_password.encode('utf-8')[:max_length], trunc)

    def test_verify_long_password_strict_raises_exception(self):
        self.config_fixture.config(strict_password_check=True)
        self.config_fixture.config(group='identity', max_password_length=5)
        invalid_password = 'passw0rd'
        self.assertRaises(
            exception.PasswordVerificationError,
            common_utils.verify_length_and_trunc_password,
            invalid_password,
        )

    def test_verify_length_and_trunc_password_throws_validation_error(self):
        class SpecialObject:
            pass

        special_object = SpecialObject()
        invalid_passwords = [True, special_object, 4.3, 5]
        for invalid_password in invalid_passwords:
            self.assertRaises(
                exception.ValidationError,
                common_utils.verify_length_and_trunc_password,
                invalid_password,
            )

    def test_hash_long_password_truncation(self):
        self.config_fixture.config(strict_password_check=False)
        invalid_length_password = '0' * 9999999
        hashed = common_utils.hash_password(invalid_length_password)
        self.assertTrue(
            common_utils.check_password(invalid_length_password, hashed)
        )

    def test_hash_long_password_strict(self):
        self.config_fixture.config(strict_password_check=True)
        invalid_length_password = '0' * 9999999
        self.assertRaises(
            exception.PasswordVerificationError,
            common_utils.hash_password,
            invalid_length_password,
        )

    def test_max_algo_length_truncates_password(self):
        self.config_fixture.config(strict_password_check=True)
        self.config_fixture.config(
            group='identity', password_hash_algorithm='bcrypt'
        )
        self.config_fixture.config(group='identity', max_password_length='96')
        invalid_length_password = '0' * 96
        self.assertRaises(
            exception.PasswordVerificationError,
            common_utils.hash_password,
            invalid_length_password,
        )

    def test_bcrypt_sha256_not_truncate_password(self):
        self.config_fixture.config(strict_password_check=True)
        self.config_fixture.config(
            group='identity', password_hash_algorithm='bcrypt_sha256'
        )
        password = '0' * 128
        password_verified = common_utils.verify_length_and_trunc_password(
            password
        )
        hashed = common_utils.hash_password(password)
        self.assertTrue(common_utils.check_password(password, hashed))
        self.assertEqual(password.encode('utf-8'), password_verified)

    def _create_test_user(self, password=OPTIONAL):
        user = {"name": "hthtest"}
        if password is not self.OPTIONAL:
            user['password'] = password

        return user

    def test_hash_user_password_without_password(self):
        user = self._create_test_user()
        hashed = common_utils.hash_user_password(user)
        self.assertEqual(user, hashed)

    def test_hash_user_password_with_null_password(self):
        user = self._create_test_user(password=None)
        hashed = common_utils.hash_user_password(user)
        self.assertEqual(user, hashed)

    def test_hash_user_password_with_empty_password(self):
        password = ''
        user = self._create_test_user(password=password)
        user_hashed = common_utils.hash_user_password(user)
        password_hashed = user_hashed['password']
        self.assertTrue(common_utils.check_password(password, password_hashed))

    def test_hash_edge_cases(self):
        hashed = common_utils.hash_password('secret')
        self.assertFalse(common_utils.check_password('', hashed))
        self.assertFalse(common_utils.check_password(None, hashed))

    def test_hash_unicode(self):
        password = 'Comment \xe7a va'
        wrong = 'Comment ?a va'
        hashed = common_utils.hash_password(password)
        self.assertTrue(common_utils.check_password(password, hashed))
        self.assertFalse(common_utils.check_password(wrong, hashed))

    def test_auth_str_equal(self):
        self.assertTrue(common_utils.auth_str_equal('abc123', 'abc123'))
        self.assertFalse(common_utils.auth_str_equal('a', 'aaaaa'))
        self.assertFalse(common_utils.auth_str_equal('aaaaa', 'a'))
        self.assertFalse(common_utils.auth_str_equal('ABC123', 'abc123'))

    def test_url_safe_check(self):
        base_str = 'i am safe'
        self.assertFalse(common_utils.is_not_url_safe(base_str))
        for i in common_utils.URL_RESERVED_CHARS:
            self.assertTrue(common_utils.is_not_url_safe(base_str + i))

    def test_url_safe_with_unicode_check(self):
        base_str = 'i am \xe7afe'
        self.assertFalse(common_utils.is_not_url_safe(base_str))
        for i in common_utils.URL_RESERVED_CHARS:
            self.assertTrue(common_utils.is_not_url_safe(base_str + i))

    def test_isotime_returns_microseconds_when_subsecond_is_true(self):
        time = timeutils.utcnow().replace(microsecond=500000)
        with freezegun.freeze_time(time):
            string_time = common_utils.isotime(subsecond=True)
        expected_string_ending = str(time.second) + '.000000Z'
        self.assertTrue(string_time.endswith(expected_string_ending))

    def test_isotime_returns_seconds_when_subsecond_is_false(self):
        time = timeutils.utcnow().replace(microsecond=500000)
        with freezegun.freeze_time(time):
            string_time = common_utils.isotime(subsecond=False)
        expected_string_ending = str(time.second) + 'Z'
        self.assertTrue(string_time.endswith(expected_string_ending))

    def test_isotime_rounds_microseconds_of_objects_passed_in(self):
        time = timeutils.utcnow().replace(microsecond=500000)
        string_time = common_utils.isotime(at=time, subsecond=True)
        expected_string_ending = str(time.second) + '.000000Z'
        self.assertTrue(string_time.endswith(expected_string_ending))

    def test_isotime_truncates_microseconds_of_objects_passed_in(self):
        time = timeutils.utcnow().replace(microsecond=500000)
        string_time = common_utils.isotime(at=time, subsecond=False)
        expected_string_ending = str(time.second) + 'Z'
        self.assertTrue(string_time.endswith(expected_string_ending))

    def test_get_certificate_subject_dn(self):
        cert_pem = unit.create_pem_certificate(
            unit.create_dn(
                common_name='test',
                organization_name='dev',
                locality_name='suzhou',
                state_or_province_name='jiangsu',
                country_name='cn',
                user_id='user_id',
                domain_component='test.com',
                email_address='user@test.com',
            )
        )

        dn = common_utils.get_certificate_subject_dn(cert_pem)
        self.assertEqual('test', dn.get('CN'))
        self.assertEqual('dev', dn.get('O'))
        self.assertEqual('suzhou', dn.get('L'))
        self.assertEqual('jiangsu', dn.get('ST'))
        self.assertEqual('cn', dn.get('C'))
        self.assertEqual('user_id', dn.get('UID'))
        self.assertEqual('test.com', dn.get('DC'))
        self.assertEqual('user@test.com', dn.get('emailAddress'))

    def test_get_certificate_issuer_dn(self):
        root_cert, root_key = unit.create_certificate(
            unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organization_name='fujitsu',
                organizational_unit_name='test',
                common_name='root',
            )
        )

        cert_pem = unit.create_pem_certificate(
            unit.create_dn(
                common_name='test',
                organization_name='dev',
                locality_name='suzhou',
                state_or_province_name='jiangsu',
                country_name='cn',
                user_id='user_id',
                domain_component='test.com',
                email_address='user@test.com',
            ),
            ca=root_cert,
            ca_key=root_key,
        )

        dn = common_utils.get_certificate_subject_dn(cert_pem)
        self.assertEqual('test', dn.get('CN'))
        self.assertEqual('dev', dn.get('O'))
        self.assertEqual('suzhou', dn.get('L'))
        self.assertEqual('jiangsu', dn.get('ST'))
        self.assertEqual('cn', dn.get('C'))
        self.assertEqual('user_id', dn.get('UID'))
        self.assertEqual('test.com', dn.get('DC'))
        self.assertEqual('user@test.com', dn.get('emailAddress'))

        dn = common_utils.get_certificate_issuer_dn(cert_pem)
        self.assertEqual('root', dn.get('CN'))
        self.assertEqual('fujitsu', dn.get('O'))
        self.assertEqual('kawasaki', dn.get('L'))
        self.assertEqual('kanagawa', dn.get('ST'))
        self.assertEqual('jp', dn.get('C'))
        self.assertEqual('test', dn.get('OU'))

    def test_get_certificate_subject_dn_not_pem_format(self):
        self.assertRaises(
            exception.ValidationError,
            common_utils.get_certificate_subject_dn,
            'MIIEkTCCAnkCFDIzsgpdRGF//5ukMuueXnRxQALhMA0GCSqGSIb3DQEBCwUAMIGC',
        )

    def test_get_certificate_issuer_dn_not_pem_format(self):
        self.assertRaises(
            exception.ValidationError,
            common_utils.get_certificate_issuer_dn,
            'MIIEkTCCAnkCFDIzsgpdRGF//5ukMuueXnRxQALhMA0GCSqGSIb3DQEBCwUAMIGC',
        )

    def test_get_certificate_thumbprint(self):
        cert_pem = '''-----BEGIN CERTIFICATE-----
        MIIEkTCCAnkCFDIzsgpdRGF//5ukMuueXnRxQALhMA0GCSqGSIb3DQEBCwUAMIGC
        MQswCQYDVQQGEwJjbjEQMA4GA1UECAwHamlhbmdzdTEPMA0GA1UEBwwGc3V6aG91
        MQ0wCwYDVQQKDARqZnR0MQwwCgYDVQQLDANkZXYxEzARBgNVBAMMCnJvb3QubG9j
        YWwxHjAcBgkqhkiG9w0BCQEWD3Rlc3RAcm9vdC5sb2NhbDAeFw0yMjA2MTYwNzM3
        NTZaFw0yMjEyMTMwNzM3NTZaMIGGMQswCQYDVQQGEwJjbjEQMA4GA1UECAwHamlh
        bmdzdTEPMA0GA1UEBwwGc3V6aG91MQ0wCwYDVQQKDARqZnR0MQwwCgYDVQQLDANk
        ZXYxFTATBgNVBAMMDGNsaWVudC5sb2NhbDEgMB4GCSqGSIb3DQEJARYRdGVzdEBj
        bGllbnQubG9jYWwwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCah1Uz
        2OVbk8zLslxxGV+AR6FTy9b/VoinmB6A0jJA1Zz2D6rsjN2S5xQ5wHIO2WSVX9Ry
        SonOmeZZqRA9faNJcNNcrBhJICScAhMGHCuli3EUMry/6xK0OYHGgI2X6mcTaIjv
        tFKHO1BCb5YGdNBa+ff+ncTeVX/PeN3nKjA4xvQb9JZxJTgY0JVhledbaoepFSdW
        EFW0nbUF+8lj1gCo5E4cAX1eTcUKs43FnWGCJcJT6FB1vP9x8e4h9p0RWbb9GMrU
        DXKbzF5e28qIiCkYHv2/A/G/J+aeg2K4Cbqy+8908I5BdWZEsJBhWJ0+CEtC3n91
        fU6dnAyipO496aa/AgMBAAEwDQYJKoZIhvcNAQELBQADggIBABoOOmLrWNlQzodS
        n2wfkiF0Lz+pj3FKFPz3sYUYWkAiKXU/6RRu1Md7INRo0MFau4iAN8Raq4JFdbnU
        HRN9G/UU58ETqi/8cYfOA2+MHHRif1Al9YSvTgHQa6ljZPttGeigOqmGlovPd+7R
        vLXlKtcr5XBVk9pWPmVpwtAN3bMVlphgEqBO26Ff9J3G5PaNQ6UdpwXC19mRqk6r
        BUsFBRwy7EeeGNy8DvoHTJfMc2JUbLjesSMOmIkaOGbhe327iRd/GJe4dO91+prE
        HNWVR/bVoGiUZvSLPqrwU173XbdNd6yMKC+fULICI34eaWDe1zHrg9XdRxtessUx
        OyJw5bgH09lOs8DSYXjFyx5lDxtERKHaLRgpSNd5foQO/mHiegC2qmdtxqKyOwub
        V/h6vziDsFZfciwmo6iw3ZpdBvjbYqw32joURQ1IVh1naY6ZzMwq/PsyYVhMYUNB
        XYPKvm68YfKuYmpwF7Z5Wll4EWm5DTq1dbmjdo+OQsMyiwWepWE0WV7Ng+AEbTqP
        /akzUXt/AEbbBpZskB6v5q/YOcglWuAQVXs2viguyDvOQVbEB7JKDi4xzlZg3kQP
        apjt17fip7wQi2jJkwdyAqvrdi/xLhK5+6BSo04lNc8sGZ9wToIoNkgv0cG+BrVU
        4cJHNiTQl8bxfSgwemgSYnnyXM4k
        -----END CERTIFICATE-----'''
        thumbprint = common_utils.get_certificate_thumbprint(cert_pem)
        self.assertEqual(
            'dMmoJKE9MIJK9VcyahYCb417JDhDfdtTiq_krco8-tk=', thumbprint
        )


class ServiceHelperTests(unit.BaseTestCase):
    @application.fail_gracefully
    def _do_test(self):
        raise Exception("Test Exc")

    def test_fail_gracefully(self):
        self.assertRaises(unit.UnexpectedExit, self._do_test)


class FernetUtilsTestCase(unit.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.config_fixture = self.useFixture(config_fixture.Config(CONF))

    def test_debug_message_logged_when_loading_fernet_token_keys(self):
        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'fernet_tokens',
                CONF.fernet_tokens.max_active_keys,
            )
        )
        logging_fixture = self.useFixture(fixtures.FakeLogger(level=log.DEBUG))
        fernet_utilities = fernet_utils.FernetUtils(
            CONF.fernet_tokens.key_repository,
            CONF.fernet_tokens.max_active_keys,
            'fernet_tokens',
        )
        fernet_utilities.load_keys()
        expected_debug_message = (
            'Loaded 2 Fernet keys from %(dir)s, but `[fernet_tokens] '
            'max_active_keys = %(max)d`; perhaps there have not been enough '
            'key rotations to reach `max_active_keys` yet?'
        ) % {
            'dir': CONF.fernet_tokens.key_repository,
            'max': CONF.fernet_tokens.max_active_keys,
        }
        self.assertIn(expected_debug_message, logging_fixture.output)

    def test_debug_message_not_logged_when_loading_fernet_credential_key(self):
        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'credential',
                CONF.fernet_tokens.max_active_keys,
            )
        )
        logging_fixture = self.useFixture(fixtures.FakeLogger(level=log.DEBUG))
        fernet_utilities = fernet_utils.FernetUtils(
            CONF.credential.key_repository,
            credential_fernet.MAX_ACTIVE_KEYS,
            'credential',
        )
        fernet_utilities.load_keys()
        debug_message = (
            'Loaded 2 Fernet keys from %(dir)s, but `[credential] '
            'max_active_keys = %(max)d`; perhaps there have not been enough '
            'key rotations to reach `max_active_keys` yet?'
        ) % {
            'dir': CONF.credential.key_repository,
            'max': credential_fernet.MAX_ACTIVE_KEYS,
        }
        self.assertNotIn(debug_message, logging_fixture.output)



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from keystone import exception
from keystone.tests import unit
from keystone.tests.unit.policy import test_backends as policy_tests


class RulesPolicy(unit.TestCase, policy_tests.PolicyTests):
    def setUp(self):
        super().setUp()
        self.load_backends()

    def config_overrides(self):
        super().config_overrides()
        self.config_fixture.config(group='policy', driver='rules')

    def test_create(self):
        self.assertRaises(exception.NotImplemented, super().test_create)

    def test_get(self):
        self.assertRaises(exception.NotImplemented, super().test_get)

    def test_list(self):
        self.assertRaises(exception.NotImplemented, super().test_list)

    def test_update(self):
        self.assertRaises(exception.NotImplemented, super().test_update)

    def test_delete(self):
        self.assertRaises(exception.NotImplemented, super().test_delete)

    def test_get_policy_returns_not_found(self):
        self.assertRaises(
            exception.NotImplemented, super().test_get_policy_returns_not_found
        )

    def test_update_policy_returns_not_found(self):
        self.assertRaises(
            exception.NotImplemented,
            super().test_update_policy_returns_not_found,
        )

    def test_delete_policy_returns_not_found(self):
        self.assertRaises(
            exception.NotImplemented,
            super().test_delete_policy_returns_not_found,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystone.common import provider_api
from keystone import exception
from keystone.tests import unit

PROVIDERS = provider_api.ProviderAPIs


class PolicyTests:
    def test_create(self):
        ref = unit.new_policy_ref()
        res = PROVIDERS.policy_api.create_policy(ref['id'], ref)
        self.assertDictEqual(ref, res)

    def test_get(self):
        ref = unit.new_policy_ref()
        res = PROVIDERS.policy_api.create_policy(ref['id'], ref)

        res = PROVIDERS.policy_api.get_policy(ref['id'])
        self.assertDictEqual(ref, res)

    def test_list(self):
        ref = unit.new_policy_ref()
        PROVIDERS.policy_api.create_policy(ref['id'], ref)

        res = PROVIDERS.policy_api.list_policies()
        res = [x for x in res if x['id'] == ref['id']][0]
        self.assertDictEqual(ref, res)

    def test_update(self):
        ref = unit.new_policy_ref()
        PROVIDERS.policy_api.create_policy(ref['id'], ref)
        orig = ref

        ref = unit.new_policy_ref()

        # (cannot change policy ID)
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.policy_api.update_policy,
            orig['id'],
            ref,
        )

        ref['id'] = orig['id']
        res = PROVIDERS.policy_api.update_policy(orig['id'], ref)
        self.assertDictEqual(ref, res)

    def test_delete(self):
        ref = unit.new_policy_ref()
        PROVIDERS.policy_api.create_policy(ref['id'], ref)

        PROVIDERS.policy_api.delete_policy(ref['id'])
        self.assertRaises(
            exception.PolicyNotFound,
            PROVIDERS.policy_api.delete_policy,
            ref['id'],
        )
        self.assertRaises(
            exception.PolicyNotFound,
            PROVIDERS.policy_api.get_policy,
            ref['id'],
        )
        res = PROVIDERS.policy_api.list_policies()
        self.assertFalse(len([x for x in res if x['id'] == ref['id']]))

    def test_get_policy_returns_not_found(self):
        self.assertRaises(
            exception.PolicyNotFound,
            PROVIDERS.policy_api.get_policy,
            uuid.uuid4().hex,
        )

    def test_update_policy_returns_not_found(self):
        ref = unit.new_policy_ref()
        self.assertRaises(
            exception.PolicyNotFound,
            PROVIDERS.policy_api.update_policy,
            ref['id'],
            ref,
        )

    def test_delete_policy_returns_not_found(self):
        self.assertRaises(
            exception.PolicyNotFound,
            PROVIDERS.policy_api.delete_policy,
            uuid.uuid4().hex,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_context import context as oslo_context

REQUEST_CONTEXT_ENV = 'keystone.oslo_request_context'


def _prop(name):
    return property(
        lambda x: getattr(x, name), lambda x, y: setattr(x, name, y)
    )


class RequestContext(oslo_context.RequestContext):
    def __init__(self, **kwargs):
        self.username = kwargs.pop('username', None)
        self.project_tag_name = kwargs.pop('project_tag_name', None)

        self.is_delegated_auth = kwargs.pop('is_delegated_auth', False)

        self.trust_id = kwargs.pop('trust_id', None)
        self.trustor_id = kwargs.pop('trustor_id', None)
        self.trustee_id = kwargs.pop('trustee_id', None)

        self.oauth_consumer_id = kwargs.pop('oauth_consumer_id', None)
        self.oauth_access_token_id = kwargs.pop('oauth_access_token_id', None)

        self.authenticated = kwargs.pop('authenticated', False)
        super().__init__(**kwargs)

    def to_policy_values(self):
        """Add keystone-specific policy values to policy representation.

        This method converts generic policy values to a dictionary form using
        the base implementation from oslo_context.context.RequestContext.
        Afterwards, it is going to pull keystone-specific values off the
        context and represent them as items in the policy values dictionary.
        This is because keystone uses default policies that rely on these
        values, so we need to guarantee they are present during policy
        enforcement if they are present on the context object.

        This method is automatically called in
        oslo_policy.policy.Enforcer.enforce() if oslo.policy knows it's dealing
        with a context object.

        """
        # TODO(morgan): Rework this to not need an explicit token render as
        # this is a generally poorly designed behavior. The enforcer should not
        # rely on a contract of the token's rendered JSON form. This likely
        # needs reworking of how we handle the context in oslo.policy. Until
        # this is reworked, it is not possible to merge the token render
        # function into keystone.api
        values = super().to_policy_values()
        values['token'] = self.token_reference['token']
        values['domain_id'] = self.domain_id if self.domain_id else None
        return values



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

DEPRECATED_REASON = """
The identity:revocation_list policy isn't used to protect any APIs in keystone
now that the revocation list API has been deprecated and only returns a 410 or
403 depending on how keystone is configured. This policy can be safely removed
from policy files.
"""

token_revocation_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'revocation_list',
        check_str=base.RULE_SERVICE_OR_ADMIN,
        # NOTE(lbragstad): Documenting scope_types here doesn't really make a
        # difference since this API is going to return an empty list regardless
        # of the token scope used in the API call. More-or-less just doing this
        # for consistency with other policies.
        scope_types=['system', 'project'],
        description='List revoked PKI tokens.',
        operations=[
            {'path': '/v3/auth/tokens/OS-PKI/revoked', 'method': 'GET'}
        ],
        deprecated_for_removal=True,
        deprecated_since=versionutils.deprecated.TRAIN,
        deprecated_reason=DEPRECATED_REASON,
    )
]


def list_rules():
    return token_revocation_policies
